# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for lookups of items in collections."""

from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, assert_never, override
from unittest import mock

from django.contrib.auth.models import AnonymousUser
from django.db.models import Q

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebusinePromise,
)
from debusine.client.models import LookupChildType
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    Group,
    User,
    WorkflowTemplate,
    Workspace,
    default_workspace,
)
from debusine.db.models.lookups import (
    LookupCacheGroup,
    LookupResult,
    LookupResultArtifact,
    LookupResultBare,
    LookupResultCollection,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections.lookup import (
    artifact_ids,
    lookup_multiple,
    lookup_single,
)
from debusine.server.collections.tests.test_base import DebusineTestManager
from debusine.tasks.models import LookupDict, LookupMultiple, LookupSingle
from debusine.test.django import TestCase


class LookupMixin(TestCase):
    """Helpers for lookup tests."""

    user: ClassVar[User]
    owners: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.owners = cls.playground.create_group(
            name="Owners", users=[cls.user]
        )

    def create_artifact_item(
        self,
        parent_collection: Collection,
        name: str,
        *,
        category: ArtifactCategory = ArtifactCategory.TEST,
        artifact: Artifact | None = None,
        data: dict[str, Any] | None = None,
    ) -> CollectionItem:
        """Create a collection item holding an artifact."""
        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            child_type=CollectionItem.Types.ARTIFACT,
            category=category,
            artifact=(
                artifact
                or self.playground.create_artifact(category=category)[0]
            ),
            data=data or {},
            created_by_user=self.user,
        )

    def create_collection_item(
        self,
        parent_collection: Collection,
        name: str,
        *,
        category: CollectionCategory = CollectionCategory.TEST,
        collection: Collection | None = None,
        data: dict[str, Any] | None = None,
    ) -> CollectionItem:
        """Create a collection item holding a collection."""
        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            child_type=CollectionItem.Types.COLLECTION,
            category=category,
            collection=(
                collection or self.playground.create_collection(name, category)
            ),
            data=data or {},
            created_by_user=self.user,
        )

    def create_source_package(self, name: str, version: str) -> Artifact:
        """Create a minimal `debian:source-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": name,
                "version": version,
                "type": "dpkg",
                "dsc_fields": {},
            },
        )
        return artifact

    def create_binary_package(
        self, name: str, version: str, architecture: str
    ) -> Artifact:
        """Create a minimal `debian:binary-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": name,
                "srcpkg_version": version,
                "deb_fields": {
                    "Package": name,
                    "Version": version,
                    "Architecture": architecture,
                },
                "deb_control_files": [],
            },
        )
        return artifact

    @staticmethod
    def make_result(
        obj: Artifact | Collection | CollectionItem,
        lookup: LookupSingle | LookupDict,
        parent_collection_lookup: LookupSingle | None = None,
    ) -> LookupResult:
        """Make a LookupResult from an underlying object."""
        match obj:
            case Artifact():
                return LookupResultArtifact(lookup=lookup, artifact=obj)
            case Collection():
                return LookupResultCollection(lookup=lookup, collection=obj)
            case CollectionItem():
                match CollectionItem.Types(obj.child_type):
                    case CollectionItem.Types.ARTIFACT:
                        return LookupResultArtifact(
                            lookup=lookup,
                            parent_collection_lookup=(
                                parent_collection_lookup
                                or f"{obj.parent_collection.id}@collections"
                            ),
                            collection_item=obj,
                            artifact=obj.artifact,
                        )
                    case CollectionItem.Types.COLLECTION:
                        return LookupResultCollection(
                            lookup=lookup,
                            parent_collection_lookup=(
                                parent_collection_lookup
                                or f"{obj.parent_collection.id}@collections"
                            ),
                            collection_item=obj,
                            collection=obj.collection,
                        )
                    case CollectionItem.Types.BARE:
                        return LookupResultBare(
                            lookup=lookup,
                            parent_collection_lookup=(
                                parent_collection_lookup
                                or f"{obj.parent_collection.id}@collections"
                            ),
                            collection_item=obj,
                        )
                    case _ as unreachable:
                        assert_never(unreachable)
            case _ as unreachable:
                assert_never(unreachable)


class LookupSingleTests(LookupMixin, TestCase):
    """Tests for lookup_single."""

    def assert_lookup_equal(
        self,
        lookup: int | str,
        expected: Artifact | Collection | CollectionItem,
        *,
        pc: PermissionContext | None = None,
        workspace: Workspace | None = None,
        parent_collection_lookup: LookupSingle | None = None,
        expect_type: LookupChildType = LookupChildType.ANY,
        **kwargs: Any,
    ) -> LookupResult:
        """Assert that a lookup's result is as expected."""
        result = lookup_single(
            lookup,
            workspace or default_workspace(),
            pc=pc or PermissionContext.for_ui_user(self.user),
            expect_type=expect_type,
            **kwargs,
        )
        assert isinstance(result, LookupResult)
        self.assertEqual(
            result,
            self.make_result(
                expected,
                lookup=lookup,
                parent_collection_lookup=parent_collection_lookup,
            ),
        )
        return result

    def assert_lookup_fails(
        self,
        lookup: int | str,
        *,
        pc: PermissionContext | None = None,
        workspace: Workspace | None = None,
        **kwargs: Any,
    ) -> None:
        """Assert that a lookup fails."""
        with self.assertRaisesRegex(
            KeyError, f"'{lookup}' does not exist or is hidden"
        ):
            lookup_single(
                lookup,
                workspace or default_workspace(),
                pc=pc or PermissionContext.for_ui_user(self.user),
                **kwargs,
            )

    def test_empty_lookup(self) -> None:
        """An empty lookup raises LookupError."""
        with self.assertRaisesRegex(LookupError, "Empty lookup"):
            lookup_single(
                "", default_workspace(), pc=PermissionContext.for_ui_anonymous()
            )

    def test_specific_artifact(self) -> None:
        """`nnn@artifacts` finds a specific artifact."""
        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]

        for artifact in artifacts:
            lookup = f"{artifact.id}@artifacts"
            with self.subTest(lookup=lookup):
                result = self.assert_lookup_equal(lookup, artifact)
                self.assertIsInstance(result, LookupResultArtifact)

    def test_specific_artifact_integer(self) -> None:
        """Integer lookups find a specific artifact."""
        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]

        for artifact in artifacts:
            with self.subTest(artifact=repr(artifact)):
                self.assert_lookup_equal(
                    artifact.id,
                    artifact,
                    expect_type=LookupChildType.ARTIFACT,
                )
                with self.assertRaisesRegex(
                    LookupError,
                    "Integer lookups only work in contexts that expect an "
                    "artifact or a collection",
                ):
                    lookup_single(
                        artifact.id,
                        default_workspace(),
                        pc=PermissionContext.for_ui_anonymous(),
                    )

    def test_specific_artifact_restricts_workspace(self) -> None:
        """`nnn@artifacts` requires the given workspace or a public one."""
        private_workspaces = []
        for i in range(2):
            workspace = self.playground.create_workspace(name=f"private{i}")
            self.owners.assign_role(workspace, "owner")
            private_workspaces.append(workspace)
        public_workspace = self.playground.create_workspace(
            name="public", public=True
        )
        artifacts = [
            self.playground.create_artifact(workspace=workspace)[0]
            for workspace in [*private_workspaces, public_workspace]
        ]

        for artifact, workspace, visible in (
            (artifacts[0], private_workspaces[0], True),
            (artifacts[0], private_workspaces[1], False),
            (artifacts[0], public_workspace, False),
            (artifacts[1], private_workspaces[0], False),
            (artifacts[1], private_workspaces[1], True),
            (artifacts[1], public_workspace, False),
            (artifacts[2], private_workspaces[0], True),
            (artifacts[2], private_workspaces[1], True),
            (artifacts[2], public_workspace, True),
        ):
            lookup = f"{artifact.id}@artifacts"
            with self.subTest(lookup=lookup, workspace=repr(workspace)):
                if visible:
                    self.assert_lookup_equal(
                        lookup, artifact, workspace=workspace
                    )
                else:
                    with self.assertRaisesRegex(
                        KeyError, f"'{lookup}' does not exist or is hidden"
                    ):
                        lookup_single(
                            lookup,
                            workspace,
                            pc=PermissionContext.for_ui_anonymous(),
                        )

    def test_specific_collection(self) -> None:
        """`nnn@collections` finds a specific collection."""
        collections = [
            self.playground.create_collection(str(i), CollectionCategory.TEST)
            for i in range(3)
        ]

        for collection in collections:
            lookup = f"{collection.id}@collections"
            with self.subTest(lookup=lookup):
                result = self.assert_lookup_equal(lookup, collection)
                self.assertIsInstance(result, LookupResultCollection)

    def test_specific_collection_integer(self) -> None:
        """Integer lookups find a specific collection."""
        collections = [
            self.playground.create_collection(str(i), CollectionCategory.TEST)
            for i in range(3)
        ]

        for collection in collections:
            with self.subTest(collection=repr(collection)):
                self.assert_lookup_equal(
                    collection.id,
                    collection,
                    expect_type=LookupChildType.COLLECTION,
                )
                with self.assertRaisesRegex(
                    LookupError,
                    "Integer lookups only work in contexts that expect an "
                    "artifact or a collection",
                ):
                    lookup_single(
                        collection.id,
                        default_workspace(),
                        pc=PermissionContext.for_ui_anonymous(),
                    )

    def test_specific_collection_restricts_workspace(self) -> None:
        """`nnn@collections` requires the given workspace or a public one."""
        private_workspaces = []
        for i in range(2):
            workspace = self.playground.create_workspace(name=f"private{i}")
            self.owners.assign_role(workspace, "owner")
            private_workspaces.append(workspace)
        public_workspace = self.playground.create_workspace(
            name="public", public=True
        )
        collections = [
            self.playground.create_collection(
                str(i), CollectionCategory.TEST, workspace=workspace
            )
            for i, workspace in enumerate(
                [*private_workspaces, public_workspace]
            )
        ]

        for collection, workspace, visible in (
            (collections[0], private_workspaces[0], True),
            (collections[0], private_workspaces[1], False),
            (collections[0], public_workspace, False),
            (collections[1], private_workspaces[0], False),
            (collections[1], private_workspaces[1], True),
            (collections[1], public_workspace, False),
            (collections[2], private_workspaces[0], True),
            (collections[2], private_workspaces[1], True),
            (collections[2], public_workspace, True),
        ):
            lookup = f"{collection.id}@collections"
            with self.subTest(lookup=lookup, workspace=repr(workspace)):
                if visible:
                    self.assert_lookup_equal(
                        lookup, collection, workspace=workspace
                    )
                else:
                    with self.assertRaisesRegex(
                        KeyError, f"'{lookup}' does not exist or is hidden"
                    ):
                        lookup_single(
                            lookup,
                            workspace,
                            pc=PermissionContext.for_ui_anonymous(),
                        )

    def test_user_restrictions(self) -> None:
        """`name@collections` respects user restrictions."""
        wpublic = self.playground.create_workspace(name="public", public=True)
        cpublic = self.playground.create_collection(
            "test", CollectionCategory.TEST, workspace=wpublic
        )
        wprivate = self.playground.create_workspace(
            name="private", public=False
        )
        self.owners.assign_role(wprivate, "owner")
        cprivate = self.playground.create_collection(
            "test", CollectionCategory.TEST, workspace=wprivate
        )
        wstart = self.playground.create_workspace(name="start", public=True)

        lookup = f"test@{CollectionCategory.TEST}"

        self.assert_lookup_fails(lookup, workspace=wstart)

        # Lookups in the workspace itself do check restrictions
        self.assert_lookup_equal(
            lookup,
            cpublic,
            pc=PermissionContext.for_ui_anonymous(),
            workspace=wpublic,
        )
        self.assert_lookup_equal(
            lookup,
            cpublic,
            pc=PermissionContext.for_ui_user(self.user),
            workspace=wpublic,
        )
        self.assert_lookup_fails(
            lookup, pc=PermissionContext.for_ui_anonymous(), workspace=wprivate
        )
        self.assert_lookup_equal(
            lookup,
            cprivate,
            pc=PermissionContext.for_ui_user(self.user),
            workspace=wprivate,
        )

        # Inheritance chain is always followed for public datasets
        wstart.set_inheritance([wpublic])
        context.clear_lookup_cache()
        self.assert_lookup_equal(
            lookup,
            cpublic,
            pc=PermissionContext.for_ui_anonymous(),
            workspace=wstart,
        )
        self.assert_lookup_equal(
            lookup,
            cpublic,
            pc=PermissionContext.for_ui_user(self.user),
            workspace=wstart,
        )

        # Inheritance chain on private datasets is followed only if logged in
        wstart.set_inheritance([wprivate])
        context.clear_lookup_cache()
        self.assert_lookup_fails(
            lookup, pc=PermissionContext.for_ui_anonymous(), workspace=wstart
        )
        self.assert_lookup_equal(
            lookup,
            cprivate,
            pc=PermissionContext.for_ui_user(self.user),
            workspace=wstart,
        )

        # Inheritance chain skips private datasets but can see public ones
        wstart.set_inheritance([wprivate, wpublic])
        context.clear_lookup_cache()
        self.assert_lookup_equal(
            lookup,
            cpublic,
            pc=PermissionContext.for_ui_anonymous(),
            workspace=wstart,
        )
        self.assert_lookup_equal(
            lookup,
            cprivate,
            pc=PermissionContext.for_ui_user(self.user),
            workspace=wstart,
        )

    def test_internal_collection_not_in_workflow(self) -> None:
        """`internal@collections` is only valid in a workflow context."""
        with self.assertRaisesRegex(
            LookupError,
            "internal@collections is only valid in the context of a workflow",
        ):
            lookup_single(
                "internal@collections",
                default_workspace(),
                pc=PermissionContext.for_ui_anonymous(),
            )

    def test_internal_collection(self) -> None:
        """`internal@collections` works in a workflow context."""
        template = WorkflowTemplate.objects.create(
            name="test", workspace=default_workspace(), task_name="noop"
        )
        workflow_root = self.playground.create_workflow(template, task_data={})
        self.playground.create_collection("test", CollectionCategory.TEST)

        assert workflow_root.internal_collection is not None
        self.assert_lookup_equal(
            "internal@collections",
            workflow_root.internal_collection,
            workflow_root=workflow_root,
        )

    def test_explicit_category(self) -> None:
        """Collections can be looked up using an explicit category."""
        collection = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        self.playground.create_collection("trixie", CollectionCategory.SUITE)
        self.playground.create_collection("debian", CollectionCategory.TEST)

        self.assert_lookup_equal(
            f"bookworm@{CollectionCategory.SUITE}", collection
        )

    def test_explicit_category_restricts_workspace(self) -> None:
        """Collection lookups require the given workspace or a public one."""
        private_workspaces: list[Workspace] = []
        for i in range(2):
            workspace = self.playground.create_workspace(name=f"private{i}")
            self.owners.assign_role(workspace, "owner")
            private_workspaces.append(workspace)
        public_workspace = self.playground.create_workspace(
            name="public", public=True
        )
        collections = [
            self.playground.create_collection(
                workspace.name, CollectionCategory.TEST, workspace=workspace
            )
            for workspace in [*private_workspaces, public_workspace]
        ]
        for workspace in private_workspaces:
            workspace.set_inheritance([public_workspace])

        for collection, workspace, visible in (
            (collections[0], private_workspaces[0], True),
            (collections[0], private_workspaces[1], False),
            (collections[0], public_workspace, False),
            (collections[1], private_workspaces[0], False),
            (collections[1], private_workspaces[1], True),
            (collections[1], public_workspace, False),
            (collections[2], private_workspaces[0], True),
            (collections[2], private_workspaces[1], True),
            (collections[2], public_workspace, True),
        ):
            lookup = f"{collection.name}@{CollectionCategory.TEST}"
            with self.subTest(lookup=lookup, workspace=repr(workspace)):
                if visible:
                    self.assert_lookup_equal(
                        lookup, collection, workspace=workspace
                    )
                else:
                    with self.assertRaisesRegex(
                        KeyError, f"'{lookup}' does not exist or is hidden"
                    ):
                        lookup_single(
                            lookup,
                            workspace,
                            pc=PermissionContext.for_ui_anonymous(),
                        )

    def test_implicit_category(self) -> None:
        """Collections can be looked up using an implicit category."""
        collection = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        self.playground.create_collection("trixie", CollectionCategory.SUITE)
        self.playground.create_collection("bookworm", CollectionCategory.TEST)

        self.assert_lookup_equal(
            "bookworm", collection, default_category=CollectionCategory.SUITE
        )

    def test_no_default_category(self) -> None:
        """An initial segment with no `@` requires a default category."""
        with self.assertRaisesRegex(
            LookupError,
            "'debian' does not specify a category and the context does not "
            "supply a default",
        ):
            lookup_single(
                "debian",
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_collection_member(self) -> None:
        """Looking up collection members works."""
        self.playground.create_collection("bullseye", CollectionCategory.SUITE)
        bookworm = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        trixie = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        src1 = self.create_source_package("src1", "1.0")
        src2 = self.create_source_package("src2", "2.0")
        items = []
        for suite, source_package_artifact in (
            (bookworm, src1),
            (bookworm, src2),
            (trixie, src2),
        ):
            items.append(
                suite.manager.add_artifact(
                    source_package_artifact,
                    user=self.user,
                    variables={"component": "main", "section": "devel"},
                )
            )

        self.assert_lookup_equal(
            "bookworm@debian:suite/source:src1",
            items[0],
            parent_collection_lookup="bookworm@debian:suite",
        )
        self.assert_lookup_equal(
            "bookworm@debian:suite/src2_2.0",
            items[1],
            parent_collection_lookup="bookworm@debian:suite",
        )
        with self.assertRaisesRegex(
            KeyError, "'bullseye@debian:suite' has no item 'source:src1'"
        ):
            lookup_single(
                "bullseye@debian:suite/source:src1",
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_no_lookup_through_artifact(self) -> None:
        """Lookup segments cannot continue after an artifact."""
        suite = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        source_package_artifact = self.create_source_package("src1", "1.0")
        suite.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )

        with self.assertRaisesRegex(
            LookupError,
            "'bookworm@debian:suite/src1_1.0' is of type 'artifact'"
            " instead of expected 'collection'",
        ):
            lookup_single(
                "bookworm@debian:suite/src1_1.0/foo",
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_expect_type(self) -> None:
        """`expect_type` requires the result to be of that type."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.WORKFLOW_INTERNAL
        )
        collection_lookup = f"collection@{CollectionCategory.WORKFLOW_INTERNAL}"
        bare_item = self.playground.create_bare_data_item(collection, "bare")
        artifact_item = self.create_artifact_item(collection, "artifact")
        collection_item = self.create_collection_item(
            collection, "sub-collection"
        )

        for name, expect_type, expected in (
            ("bare", LookupChildType.BARE, bare_item),
            ("bare", LookupChildType.ARTIFACT, None),
            ("bare", LookupChildType.COLLECTION, None),
            ("bare", LookupChildType.ANY, bare_item),
            ("artifact", LookupChildType.BARE, None),
            ("artifact", LookupChildType.ARTIFACT, artifact_item),
            ("artifact", LookupChildType.COLLECTION, None),
            ("artifact", LookupChildType.ANY, artifact_item),
            ("sub-collection", LookupChildType.BARE, None),
            ("sub-collection", LookupChildType.ARTIFACT, None),
            ("sub-collection", LookupChildType.COLLECTION, collection_item),
            ("sub-collection", LookupChildType.ANY, collection_item),
        ):
            lookup = f"{collection_lookup}/{name}"
            with self.subTest(lookup=lookup, expect_type=expect_type.value):
                if expected is not None:
                    self.assert_lookup_equal(
                        lookup,
                        expected,
                        expect_type=expect_type,
                        parent_collection_lookup=collection_lookup,
                    )
                else:
                    with self.assertRaisesRegex(
                        LookupError,
                        f"{lookup!r} is of type '.*'"
                        f" instead of expected {expect_type.name.lower()!r}",
                    ):
                        lookup_single(
                            lookup,
                            default_workspace(),
                            pc=PermissionContext.for_ui_user(self.user),
                            expect_type=expect_type,
                        )

    def _add_parent_workspace(self, other_scope: bool = False) -> None:
        """Add inheritance with some collections hiding each other"""
        workspace = default_workspace()
        scope = workspace.scope
        if other_scope:
            scope = self.playground.get_or_create_scope(name="scope")
            self.other_scope = scope

        self.parent_workspace = self.playground.create_parent_workspace(
            child=workspace, name="workspace1", public=True, scope=scope
        )
        # Add collections
        self.playground.create_collection(
            "collection1", CollectionCategory.TEST, workspace=workspace
        )
        self.playground.create_collection(
            "collection1",
            CollectionCategory.TEST,
            workspace=self.parent_workspace,
        )

    def test_lookup_with_workspace_prefix(self) -> None:
        """The collection returned is from the requested workspace"""
        self._add_parent_workspace()
        lookup = f"//workspace=workspace1/collection1@{CollectionCategory.TEST}"
        result = lookup_single(
            lookup,
            default_workspace(),
            pc=PermissionContext.for_ui_user(self.user),
        )
        assert result.collection
        self.assertEqual(result.collection.workspace, self.parent_workspace)

    def test_lookup_with_unknown_workspace(self) -> None:
        lookup = (
            f"//workspace=doesnotexist/collection1@{CollectionCategory.TEST}"
        )
        with self.assertRaisesRegex(
            LookupError, r"^Invalid or hidden workspace in lookup options:"
        ):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_lookup_with_private_workspace(self) -> None:
        self._add_parent_workspace()
        self.parent_workspace.public = False
        self.parent_workspace.save()
        lookup = f"//workspace=workspace1/collection1@{CollectionCategory.TEST}"
        with self.assertRaisesRegex(
            LookupError, r"^Invalid or hidden workspace in lookup options:"
        ):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_lookup_with_unrelated_workspace(self) -> None:
        workspace = default_workspace()
        self.playground.create_collection(
            "collection1", CollectionCategory.TEST, workspace=workspace
        )
        unrelated_workspace = self.playground.create_workspace(
            name="unrelated", public=True
        )
        collection = self.playground.create_collection(
            "collection1",
            CollectionCategory.TEST,
            workspace=unrelated_workspace,
        )
        lookup = f"//workspace=unrelated/collection1@{CollectionCategory.TEST}"
        result = lookup_single(
            lookup,
            default_workspace(),
            pc=PermissionContext.for_ui_user(self.user),
        )
        self.assertEqual(result.collection, collection)

    def test_lookup_with_unknown_option(self) -> None:
        lookup = f"//invalid=value/collection1@{CollectionCategory.TEST}"
        with self.assertRaisesRegex(LookupError, r"^Unknown lookup option:"):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_lookup_empty_options(self) -> None:
        self.playground.create_collection(
            "collection1", CollectionCategory.TEST
        )

        lookup = f"//invalid/collection1@{CollectionCategory.TEST}"
        with self.assertRaisesRegex(
            LookupError, r"^Invalid syntax in lookup options part of"
        ):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_lookup_in_other_scope(self) -> None:
        self._add_parent_workspace(other_scope=True)
        lookup = (
            f"//workspace=workspace1:scope={self.other_scope.name}/"
            f"collection1@{CollectionCategory.TEST}"
        )
        result = lookup_single(
            lookup,
            default_workspace(),
            pc=PermissionContext.for_ui_user(self.user),
        )
        assert result.collection
        self.assertEqual(result.collection.workspace, self.parent_workspace)

    def test_lookup_in_invalid_scope(self) -> None:
        self._add_parent_workspace(other_scope=True)
        lookup = (
            f"//scope=doesnotexist:workspace=workspace1/"
            f"collection1@{CollectionCategory.TEST}"
        )
        with self.assertRaisesRegex(
            LookupError, r"^Invalid or hidden workspace in lookup options:"
        ):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_lookup_scope_without_workspace(self) -> None:
        self._add_parent_workspace(other_scope=True)
        lookup = (
            f"//scope={self.other_scope.name}/"
            f"collection1@{CollectionCategory.TEST}"
        )
        with self.assertRaisesRegex(
            LookupError,
            r"^The 'scope' lookup option requires the 'workspace' one",
        ):
            lookup_single(
                lookup,
                default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_caching(self) -> None:
        this_workspace = default_workspace()
        other_workspace = self.playground.create_workspace(name="other")
        workflow = self.playground.create_workflow()
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        item = self.create_artifact_item(collection, "artifact")
        lookup = f"{collection_lookup}/artifact"
        expected_cache: LookupCacheGroup[LookupSingle, LookupResult] = (
            LookupCacheGroup()
        )

        for (
            workspace,
            user,
            default_category,
            workflow_root,
            expect_type,
            expected_queries,
            expected,
        ) in (
            (None, None, None, None, None, 3, item),
            (other_workspace, None, None, None, None, 1, None),
            (None, AnonymousUser(), None, None, None, 3, item),
            (None, None, CollectionCategory.SUITE, None, None, 3, item),
            (None, None, None, workflow, None, 3, item),
            (None, None, None, None, LookupChildType.BARE, 3, None),
        ):
            lookup_workspace = workspace or this_workspace
            if (pc_user := user or self.user).is_authenticated:
                lookup_pc = PermissionContext.for_ui_user(pc_user)
            else:
                lookup_pc = PermissionContext.for_ui_anonymous()
            assert lookup_pc.user is not None
            lookup_expect_type = expect_type or LookupChildType.ANY
            with (
                self.subTest(
                    workspace=repr(workspace),
                    user=repr(user),
                    default_category=str(default_category),
                    workflow_root=repr(workflow_root),
                    expect_type=str(expect_type),
                ),
                self.assertNumQueries(expected_queries),
            ):
                if expected is not None:
                    result = self.assert_lookup_equal(
                        lookup,
                        expected,
                        pc=lookup_pc,
                        workspace=lookup_workspace,
                        parent_collection_lookup=collection_lookup,
                        default_category=default_category,
                        workflow_root=workflow_root,
                        expect_type=lookup_expect_type,
                    )
                    expected_cache[
                        (
                            lookup,
                            lookup_workspace,
                            lookup_pc.user,
                            default_category,
                            workflow_root,
                            lookup_expect_type,
                        )
                    ] = result
                else:
                    with self.assertRaises(LookupError):
                        lookup_single(
                            lookup,
                            lookup_workspace,
                            pc=lookup_pc,
                            default_category=default_category,
                            workflow_root=workflow_root,
                            expect_type=lookup_expect_type,
                        )

        assert context.lookup_cache.single == expected_cache

        with self.assertNumQueries(0):
            self.assert_lookup_equal(
                lookup,
                item,
                workspace=this_workspace,
                parent_collection_lookup=collection_lookup,
            )

        self.assertEqual(context.lookup_cache.single, expected_cache)

    def test_precache_result(self) -> None:
        workspace = default_workspace()
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        item = self.create_artifact_item(collection, "artifact")
        lookup = f"{collection_lookup}/artifact"
        result = LookupResultArtifact(
            lookup=lookup,
            parent_collection_lookup=collection_lookup,
            collection_item=item,
            artifact=item.artifact,
        )

        with self.assertNumQueries(0):
            self.assert_lookup_equal(
                lookup,
                item,
                workspace=workspace,
                parent_collection_lookup=collection_lookup,
                precache_result=result,
            )
        with self.assertNumQueries(0):
            self.assert_lookup_equal(
                lookup,
                item,
                workspace=workspace,
                parent_collection_lookup=collection_lookup,
            )


class LookupMultipleTests(LookupMixin, TestCase):
    """Tests for lookup_multiple."""

    def parse_lookup(
        self, data: dict[str, Any] | Sequence[int | str | dict[str, Any]]
    ) -> LookupMultiple:
        """Parse the lookup syntax."""
        return LookupMultiple.model_validate(data)

    def assert_lookup_equal(
        self,
        lookup: dict[str, Any] | Sequence[int | str | dict[str, Any]],
        expected: Iterable[Artifact | Collection | CollectionItem],
        *,
        pc: PermissionContext | None = None,
        workspace: Workspace | None = None,
        parent_collection_lookup: LookupSingle | None = None,
        **kwargs: Any,
    ) -> Sequence[LookupResult]:
        """Assert that a lookup's result is as expected."""
        parsed_lookup = self.parse_lookup(lookup)
        result = lookup_multiple(
            parsed_lookup,
            workspace or default_workspace(),
            pc=pc or PermissionContext.for_ui_user(self.user),
            **kwargs,
        )
        self.assertEqual(
            result,
            tuple(
                self.make_result(
                    item,
                    # Just use the returned lookup; this is hard to predict
                    # in the general case without making tests much more
                    # verbose.
                    lookup=r.lookup,
                    parent_collection_lookup=parent_collection_lookup,
                )
                for item, r in zip(expected, result)
            ),
        )
        return result

    def assert_lookup_fails(
        self,
        lookup: dict[str, Any] | Sequence[int | str | dict[str, Any]],
        expected_message: str,
        *,
        workspace: Workspace | None = None,
        **kwargs: Any,
    ) -> None:
        """Assert that a lookup fails."""
        with self.assertRaisesRegex(KeyError, expected_message):
            lookup_multiple(
                self.parse_lookup(lookup),
                workspace or default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
                **kwargs,
            )

    def test_not_collection(self) -> None:
        """The `collection` key does not identify a collection."""
        artifact = self.playground.create_artifact()[0]
        lookup = {"collection": f"{artifact.id}@artifacts"}

        with self.assertRaisesRegex(
            LookupError,
            f"'{artifact.id}@artifacts' is of type 'artifact'"
            f" instead of expected 'collection'",
        ):
            lookup_multiple(
                self.parse_lookup(lookup),
                default_workspace(),
                pc=PermissionContext.for_ui_anonymous(),
            )

    def test_child_type_bare(self) -> None:
        """Restrict to bare items."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        other_collection = self.playground.create_collection(
            "other-collection", CollectionCategory.TEST
        )
        bare1 = self.playground.create_bare_data_item(collection, "bare1")
        bare2 = self.playground.create_bare_data_item(collection, "bare2")
        self.create_artifact_item(collection, "artifact")
        self.create_collection_item(collection, "sub-collection")
        self.playground.create_bare_data_item(other_collection, "bare3")

        result = self.assert_lookup_equal(
            {"collection": collection_lookup, "child_type": "bare"},
            (bare1, bare2),
            parent_collection_lookup=collection_lookup,
        )
        self.assertIsInstance(result[0], LookupResultBare)
        self.assertIsInstance(result[1], LookupResultBare)

    def test_child_type_artifact(self) -> None:
        """Restrict to artifacts."""
        bookworm = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        trixie = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        src1 = self.create_source_package("src1", "1.0")
        src2 = self.create_source_package("src2", "2.0")
        self.playground.create_bare_data_item(
            bookworm, "bare1", category=BareDataCategory.TEST
        )
        bare2 = self.playground.create_bare_data_item(
            bookworm,
            "bare2",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=2,
                promise_workflow_id=1,
                promise_category=ArtifactCategory.TEST,
            ),
        )
        self.create_collection_item(bookworm, "collection")
        items = []
        for suite, source_package_artifact in (
            (bookworm, src1),
            (bookworm, src2),
            (trixie, src2),
        ):
            items.append(
                suite.manager.add_artifact(
                    source_package_artifact,
                    user=self.user,
                    variables={"component": "main", "section": "devel"},
                )
            )

        self.assert_lookup_equal(
            {"collection": "bookworm@debian:suite", "child_type": "artifact"},
            items[:2],
            parent_collection_lookup="bookworm@debian:suite",
        )
        self.assert_lookup_equal(
            {
                "collection": "bookworm@debian:suite",
                "child_type": "artifact-or-promise",
            },
            [bare2, *items[:2]],
            parent_collection_lookup="bookworm@debian:suite",
        )
        self.assert_lookup_equal(
            {"collection": "bookworm@debian:suite"},
            items[:2],
            parent_collection_lookup="bookworm@debian:suite",
        )

    def test_child_type_collection(self) -> None:
        """Restrict to collections."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        other_collection = self.playground.create_collection(
            "other-collection", CollectionCategory.TEST
        )
        self.playground.create_bare_data_item(collection, "bare1")
        self.create_artifact_item(collection, "artifact")
        collection1 = self.create_collection_item(collection, "collection1")
        collection2 = self.create_collection_item(collection, "collection2")
        self.create_collection_item(other_collection, "collection3")

        self.assert_lookup_equal(
            {"collection": collection_lookup, "child_type": "collection"},
            (collection1, collection2),
            parent_collection_lookup=collection_lookup,
        )

    def test_child_type_any(self) -> None:
        """Don't restrict by child type."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        bare = self.playground.create_bare_data_item(collection, "bare")
        artifact = self.create_artifact_item(collection, "artifact")
        sub_collection = self.create_collection_item(
            collection, "sub-collection"
        )

        self.assert_lookup_equal(
            {"collection": collection_lookup, "child_type": "any"},
            (bare, artifact, sub_collection),
            parent_collection_lookup=collection_lookup,
        )

    def test_collection_lookup_support_options(self) -> None:
        """The collection lookup can have prefixed options."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        w_name = self.playground.default_workspace_name
        collection_lookup = (
            f"//workspace={w_name}/collection@{CollectionCategory.TEST}"
        )
        artifact = self.create_artifact_item(collection, "artifact")

        self.assert_lookup_equal(
            {"collection": collection_lookup, "child_type": "any"},
            (artifact,),
            parent_collection_lookup=collection_lookup,
        )

    def test_error_on_empty(self) -> None:
        """`error_on_empty: True` makes empty results an error."""
        self.playground.create_collection("collection", CollectionCategory.TEST)
        collection_lookup = f"collection@{CollectionCategory.TEST}"

        for lookup, expected in (
            ({"collection": collection_lookup}, []),
            ({"collection": collection_lookup, "error_on_empty": True}, None),
        ):
            with self.subTest(lookup=lookup):
                if expected is not None:
                    self.assert_lookup_equal(
                        lookup,
                        expected,
                        parent_collection_lookup=collection_lookup,
                    )
                else:
                    with self.assertRaisesRegex(
                        LookupError,
                        f"No results found from"
                        f" {LookupMultiple.model_validate([lookup]).export()}",
                    ):
                        lookup_multiple(
                            self.parse_lookup(lookup),
                            default_workspace(),
                            pc=PermissionContext.for_ui_user(self.user),
                        )

    def test_expect_type(self) -> None:
        """`expect_type` requires results to be of that type."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        bare_item = self.playground.create_bare_data_item(collection, "bare")
        artifact_item = self.create_artifact_item(collection, "artifact")
        collection_item = self.create_collection_item(
            collection, "sub-collection"
        )

        for child_type, expect_type, expected in (
            ("bare", LookupChildType.BARE, [bare_item]),
            ("bare", LookupChildType.ARTIFACT, None),
            ("bare", LookupChildType.ARTIFACT_OR_PROMISE, [bare_item]),
            ("bare", LookupChildType.COLLECTION, None),
            ("bare", LookupChildType.ANY, [bare_item]),
            ("artifact", LookupChildType.BARE, None),
            ("artifact", LookupChildType.ARTIFACT, [artifact_item]),
            ("artifact", LookupChildType.ARTIFACT_OR_PROMISE, [artifact_item]),
            ("artifact", LookupChildType.COLLECTION, None),
            ("artifact", LookupChildType.ANY, [artifact_item]),
            ("artifact-or-promise", LookupChildType.BARE, None),
            ("artifact-or-promise", LookupChildType.ARTIFACT, [artifact_item]),
            (
                "artifact-or-promise",
                LookupChildType.ARTIFACT_OR_PROMISE,
                [artifact_item],
            ),
            ("artifact-or-promise", LookupChildType.COLLECTION, None),
            ("artifact-or-promise", LookupChildType.ANY, [artifact_item]),
            ("collection", LookupChildType.BARE, None),
            ("collection", LookupChildType.ARTIFACT, None),
            ("collection", LookupChildType.ARTIFACT_OR_PROMISE, None),
            ("collection", LookupChildType.COLLECTION, [collection_item]),
            ("collection", LookupChildType.ANY, [collection_item]),
            ("any", LookupChildType.BARE, None),
            ("any", LookupChildType.ARTIFACT, None),
            ("any", LookupChildType.ARTIFACT_OR_PROMISE, None),
            ("any", LookupChildType.COLLECTION, None),
            (
                "any",
                LookupChildType.ANY,
                [bare_item, artifact_item, collection_item],
            ),
        ):
            lookup = {
                "collection": f"collection@{CollectionCategory.TEST}",
                "child_type": child_type,
            }
            with self.subTest(lookup=lookup, expect_type=expect_type.value):
                if expected is not None:
                    self.assert_lookup_equal(
                        lookup,
                        expected,
                        expect_type=expect_type,
                        parent_collection_lookup=collection_lookup,
                    )
                else:
                    with self.assertRaisesRegex(
                        LookupError,
                        f"Result from"
                        f" {LookupMultiple.model_validate([lookup]).export()}"
                        f" is of type '.*' instead of expected"
                        f" {expect_type.name.lower()!r}",
                    ):
                        lookup_multiple(
                            self.parse_lookup(lookup),
                            default_workspace(),
                            pc=PermissionContext.for_ui_user(self.user),
                            expect_type=expect_type,
                        )

    def test_category(self) -> None:
        """Restrict by category."""
        bookworm = self.playground.create_collection(
            "bookworm", CollectionCategory.SUITE
        )
        src1 = self.create_source_package("src1", "1.0")
        src2 = self.create_source_package("src2", "2.0")
        bin1 = self.create_binary_package("bin1", "1.0", "amd64")
        bin2 = self.create_binary_package("bin2", "2.0", "s390x")
        items = [
            bookworm.manager.add_artifact(
                source_package_artifact,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )
            for source_package_artifact in (src1, src2)
        ]
        items += [
            bookworm.manager.add_artifact(
                binary_package_artifact,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )
            for binary_package_artifact in (bin1, bin2)
        ]

        self.assert_lookup_equal(
            {
                "collection": "bookworm@debian:suite",
                "category": ArtifactCategory.SOURCE_PACKAGE,
            },
            items[:2],
            parent_collection_lookup="bookworm@debian:suite",
        )
        self.assert_lookup_equal(
            {
                "collection": "bookworm@debian:suite",
                "category": ArtifactCategory.BINARY_PACKAGE,
            },
            items[2:],
            parent_collection_lookup="bookworm@debian:suite",
        )

    def test_name_matcher(self) -> None:
        """Restrict by name."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        items = [
            self.create_artifact_item(collection, name)
            for name in ("foo", "foobar", "rebar")
        ]

        self.assert_lookup_equal(
            {"collection": collection_lookup, "name": "foo"},
            (items[0],),
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "name__startswith": "foo"},
            items[:2],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "name__endswith": "bar"},
            items[1:],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "name__contains": "oo"},
            items[:2],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "name__in": ("foo", "rebar")},
            (items[0], items[2]),
            parent_collection_lookup=collection_lookup,
        )

    def test_data_matchers(self) -> None:
        """Restrict by data."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        item_specs: list[tuple[str, dict[str, Any]]] = [
            ("foo", {"package": "foo"}),
            ("foobar_1.0", {"package": "foobar", "version": "1.0"}),
            ("rebar", {"package": "rebar", "some_id": 123}),
        ]
        items = [
            self.create_artifact_item(collection, name, data=data)
            for name, data in item_specs
        ]

        self.assert_lookup_equal(
            {"collection": collection_lookup, "data__package": "foo"},
            (items[0],),
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {
                "collection": collection_lookup,
                "data__package__startswith": "foo",
            },
            items[:2],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "data__package__endswith": "bar"},
            items[1:],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "data__package__contains": "oo"},
            items[:2],
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {
                "collection": collection_lookup,
                "data__package__in": ("foo", "rebar"),
            },
            (items[0], items[2]),
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "data__version": "1.0"},
            (items[1],),
            parent_collection_lookup=collection_lookup,
        )
        self.assert_lookup_equal(
            {"collection": collection_lookup, "data__some_id": 123},
            (items[2],),
            parent_collection_lookup=collection_lookup,
        )

    def test_lookup_filters(self) -> None:
        """Restrict by custom lookup filter."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        item_specs: list[tuple[str, dict[str, Any]]] = [
            ("foo", {"package": "foo"}),
            ("foobar_1.0", {"package": "foobar", "version": "1.0"}),
            ("rebar", {"package": "rebar", "some_id": 123}),
        ]
        items = [
            self.create_artifact_item(collection, name, data=data)
            for name, data in item_specs
        ]
        template = WorkflowTemplate.objects.create(
            name="test", workspace=default_workspace(), task_name="noop"
        )
        workflow_root = self.playground.create_workflow(template, task_data={})

        def do_lookup_filter(
            key: str, value: LookupSingle | LookupMultiple, **kwargs: Any
        ) -> Q:
            self.assertEqual(key, "test")
            self.assertIsInstance(value, LookupSingle)
            return Q(data__package=value)

        with mock.patch.object(
            DebusineTestManager,
            "do_lookup_filter",
            side_effect=do_lookup_filter,
        ) as mock_do_lookup_filter:
            self.assert_lookup_equal(
                {"collection": collection_lookup, "lookup__test": "foo"},
                (items[0],),
                parent_collection_lookup=collection_lookup,
            )
            mock_do_lookup_filter.assert_called_once_with(
                "test",
                "foo",
                workspace=default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
                workflow_root=None,
            )
            mock_do_lookup_filter.reset_mock()

            self.assert_lookup_equal(
                {"collection": collection_lookup, "lookup__test": "foobar"},
                (items[1],),
                workflow_root=workflow_root,
                parent_collection_lookup=collection_lookup,
            )
            mock_do_lookup_filter.assert_called_once_with(
                "test",
                "foobar",
                workspace=default_workspace(),
                pc=PermissionContext.for_ui_user(self.user),
                workflow_root=workflow_root,
            )

    def test_alternatives(self) -> None:
        """The list syntax allows specifying several lookups at once."""
        trixie = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        items = [
            self.create_artifact_item(trixie, name, data=data)
            for name, data in (
                (
                    "libc6-dev_2.37-15",
                    {"package": "libc6-dev", "version": "2.37-15"},
                ),
                (
                    "libc6-dev_2.37-16",
                    {"package": "libc6-dev", "version": "2.37-16"},
                ),
                (
                    "debhelper_13.15.3_amd64",
                    {
                        "package": "debhelper",
                        "version": "13.15.3",
                        "architecture": "amd64",
                    },
                ),
                (
                    "debhelper_13.15.3_s390x",
                    {
                        "package": "debhelper",
                        "version": "13.15.3",
                        "architecture": "s390x",
                    },
                ),
                ("hello_1.0", {"package": "hello", "version": "1.0"}),
            )
        ]
        loose_artifacts = [
            self.playground.create_artifact()[0] for _ in range(3)
        ]

        self.assert_lookup_equal(
            [
                {
                    "collection": "trixie",
                    "data__package": "libc6-dev",
                    "data__version": "2.37-15",
                },
                {
                    "collection": "trixie",
                    "name__startswith": "debhelper_13.15.3_",
                },
                f"{loose_artifacts[0].id}@artifacts",
                loose_artifacts[1].id,
            ],
            [
                items[0],
                items[2],
                items[3],
                loose_artifacts[0],
                loose_artifacts[1],
            ],
            default_category=CollectionCategory.SUITE,
            expect_type=LookupChildType.ARTIFACT,
            parent_collection_lookup=f"trixie@{CollectionCategory.SUITE}",
        )

    def test_artifact_expired(self) -> None:
        """If one artifact has expired, `lookup_multiple` raises LookupError."""
        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]
        artifact_ids = [artifact.id for artifact in artifacts]
        artifacts[0].delete()

        self.assert_lookup_fails(
            artifact_ids,
            f"'{artifact_ids[0]}@artifacts' does not exist or is hidden",
            expect_type=LookupChildType.ARTIFACT,
        )

    def test_collection_removed(self) -> None:
        """If one collection is gone, `lookup_multiple` raises LookupError."""
        collections = [
            self.playground.create_collection(
                f"collection{i}", CollectionCategory.TEST
            )
            for i in range(3)
        ]
        items = [
            self.create_artifact_item(collection, "artifact")
            for collection in collections
        ]
        items[0].delete()
        collections[0].delete()

        self.assert_lookup_fails(
            [
                {
                    "collection": f"collection{i}@{CollectionCategory.TEST}",
                    "name": "artifact",
                }
                for i in range(3)
            ],
            f"'collection0@{CollectionCategory.TEST}' does not exist or is "
            f"hidden",
        )

    def test_query_count(self) -> None:
        """`lookup_multiple` makes a constant number of DB queries."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        artifact_items = [
            self.create_artifact_item(collection, f"artifact{i}")
            for i in range(10)
        ]
        collection_items = [
            self.create_collection_item(collection, f"collection{i}")
            for i in range(10)
        ]

        # Each of these makes 4 queries:
        # 1. look up the workspace (in assert_lookup_equal)
        # *. check if the workspace is visible (shortcutted as it's public)
        # 2. find the containing collection
        # 3. list the set of collection items.
        with self.assertNumQueries(3):
            self.assert_lookup_equal(
                {"collection": collection_lookup},
                artifact_items,
                parent_collection_lookup=collection_lookup,
            )
        context.clear_lookup_cache()
        with self.assertNumQueries(3):
            self.assert_lookup_equal(
                {
                    "collection": collection_lookup,
                    "child_type": "collection",
                },
                collection_items,
                parent_collection_lookup=collection_lookup,
            )

    def test_caching(self) -> None:
        this_workspace = default_workspace()
        other_workspace = self.playground.create_workspace(name="other")
        workflow = self.playground.create_workflow()
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST
        )
        collection_lookup = f"collection@{CollectionCategory.TEST}"
        items = [
            self.create_artifact_item(collection, f"artifact{i}")
            for i in range(10)
        ]
        lookup = {"collection": collection_lookup}
        expected_cache: LookupCacheGroup[LookupDict, Sequence[LookupResult]] = (
            LookupCacheGroup()
        )

        for (
            workspace,
            user,
            default_category,
            workflow_root,
            expect_type,
            expected_queries,
            expected,
        ) in (
            (None, None, None, None, None, 2, items),
            (other_workspace, None, None, None, None, 1, None),
            (None, AnonymousUser(), None, None, None, 2, items),
            (None, None, CollectionCategory.SUITE, None, None, 2, items),
            (None, None, None, workflow, None, 2, items),
            (None, None, None, None, LookupChildType.BARE, 1, None),
        ):
            lookup_workspace = workspace or this_workspace
            if (pc_user := user or self.user).is_authenticated:
                lookup_pc = PermissionContext.for_ui_user(pc_user)
            else:
                lookup_pc = PermissionContext.for_ui_anonymous()
            assert lookup_pc.user is not None
            lookup_expect_type = expect_type or LookupChildType.ANY
            with (
                self.subTest(
                    workspace=repr(workspace),
                    user=repr(user),
                    default_category=str(default_category),
                    workflow_root=repr(workflow_root),
                    expect_type=str(expect_type),
                ),
                self.assertNumQueries(expected_queries),
            ):
                if expected is not None:
                    self.assert_lookup_equal(
                        lookup,
                        expected,
                        pc=lookup_pc,
                        workspace=lookup_workspace,
                        parent_collection_lookup=collection_lookup,
                        default_category=default_category,
                        workflow_root=workflow_root,
                        expect_type=lookup_expect_type,
                    )
                    expected_cache[
                        (
                            LookupDict.model_validate(lookup),
                            lookup_workspace,
                            lookup_pc.user,
                            default_category,
                            workflow_root,
                            lookup_expect_type,
                        )
                    ] = tuple(
                        self.make_result(
                            item,
                            lookup=LookupDict.model_validate(lookup),
                            parent_collection_lookup=collection_lookup,
                        )
                        for item in expected
                    )
                else:
                    with self.assertRaises(LookupError):
                        lookup_multiple(
                            self.parse_lookup(lookup),
                            lookup_workspace,
                            pc=lookup_pc,
                            default_category=default_category,
                            workflow_root=workflow_root,
                            expect_type=lookup_expect_type,
                        )

        self.assertEqual(context.lookup_cache.dictionary, expected_cache)

        with self.assertNumQueries(0):
            self.assert_lookup_equal(
                lookup,
                items,
                workspace=this_workspace,
                parent_collection_lookup=collection_lookup,
            )

        self.assertEqual(context.lookup_cache.dictionary, expected_cache)


class ReconstructLookupTests(LookupMixin, TestCase):
    """Tests for reconstructing lookups from their results."""

    @override
    def setUp(self) -> None:
        super().setUp()
        # A basic dictionary lookup that's enough to trigger the more
        # complex reconstruction code.
        self.basic_lookup_dict = LookupDict.model_validate(
            {"collection": f"test@{CollectionCategory.TEST}"}
        )

    def test_with_lookup_single(self) -> None:
        """A result that came from a single lookup returns that lookup."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        lookup = f"trixie@{CollectionCategory.SUITE}/source:hello"
        result = self.make_result(artifact, lookup=lookup)

        self.assertEqual(result.reconstruct(), lookup)

    def test_collection_item(self) -> None:
        """A result with a collection item returns a lookup for that item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        collection_lookup = f"test@{CollectionCategory.TEST}"
        item_bare = self.playground.create_bare_data_item(collection, "bare")
        item_artifact = self.create_artifact_item(collection, "artifact")
        item_collection = self.create_collection_item(collection, "collection")
        workflow = self.playground.create_workflow()

        for item, name in (
            (item_bare, "bare"),
            (item_artifact, "artifact"),
            (item_collection, "collection"),
        ):
            result = self.make_result(
                item,
                lookup=self.basic_lookup_dict,
                parent_collection_lookup=collection_lookup,
            )
            self.assertEqual(
                result.reconstruct(), f"{collection_lookup}/name:{name}"
            )
            self.assertEqual(
                result.reconstruct(workflow_root=workflow),
                f"{collection_lookup}/name:{name}",
            )

    def test_artifact(self) -> None:
        """A result with an artifact returns a lookup for that artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        result = self.make_result(artifact, lookup=self.basic_lookup_dict)

        self.assertEqual(
            result.reconstruct(),
            f"{artifact.id}@artifacts",
        )

    def test_collection(self) -> None:
        """A result with a collection returns a lookup for that collection."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        result = self.make_result(collection, lookup=self.basic_lookup_dict)

        self.assertEqual(result.reconstruct(), f"{collection.id}@collections")
        self.assertEqual(
            result.reconstruct(
                workflow_root=self.playground.create_workflow(),
            ),
            f"{collection.id}@collections",
        )

    def test_collection_internal_collection(self) -> None:
        """The internal collection is handled specially."""
        workflow = self.playground.create_workflow()
        assert workflow.internal_collection is not None
        result = self.make_result(
            workflow.internal_collection, lookup=self.basic_lookup_dict
        )

        self.assertEqual(
            result.reconstruct(workflow_root=workflow), "internal@collections"
        )


class ArtifactIdsTests(LookupMixin, TestCase):
    """Tests for artifact_ids() function."""

    def test_return_artifact_ids(self) -> None:
        artifact = self.playground.create_minimal_binary_package_artifact()
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )

        items = [
            LookupResultArtifact(lookup=artifact.id, artifact=artifact),
            LookupResultCollection(
                lookup=f"test@{CollectionCategory.TEST}", collection=collection
            ),
        ]
        self.assertEqual(artifact_ids(items), [artifact.id])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine Workflow orchestration infrastructure."""

import sys
from typing import cast

from debusine.server.workflows.autopkgtest import AutopkgtestWorkflow
from debusine.server.workflows.base import Workflow, WorkflowValidationError
from debusine.server.workflows.blhc import BlhcWorkflow
from debusine.server.workflows.create_child_workspace import (
    CreateChildWorkspaceWorkflow,
)
from debusine.server.workflows.debdiff import DebDiffWorkflow
from debusine.server.workflows.debian_pipeline import DebianPipelineWorkflow
from debusine.server.workflows.lintian import LintianWorkflow
from debusine.server.workflows.make_signed_source import (
    MakeSignedSourceWorkflow,
)
from debusine.server.workflows.noop import NoopWorkflow
from debusine.server.workflows.package_publish import PackagePublishWorkflow
from debusine.server.workflows.package_upload import PackageUploadWorkflow
from debusine.server.workflows.piuparts import PiupartsWorkflow
from debusine.server.workflows.qa import QAWorkflow
from debusine.server.workflows.reverse_dependencies_autopkgtest import (
    ReverseDependenciesAutopkgtestWorkflow,
)
from debusine.server.workflows.sbuild import SbuildWorkflow
from debusine.server.workflows.update_environments import (
    UpdateEnvironmentsWorkflow,
)
from debusine.server.workflows.update_suite import UpdateSuiteWorkflow
from debusine.server.workflows.update_suites import UpdateSuitesWorkflow

# Import the documentation from where the code lives
__doc__ = cast(str, sys.modules[Workflow.__module__].__doc__)

__all__ = [
    "AutopkgtestWorkflow",
    "BlhcWorkflow",
    "CreateChildWorkspaceWorkflow",
    "DebDiffWorkflow",
    "DebianPipelineWorkflow",
    "LintianWorkflow",
    "MakeSignedSourceWorkflow",
    "NoopWorkflow",
    "PackagePublishWorkflow",
    "PackageUploadWorkflow",
    "PiupartsWorkflow",
    "QAWorkflow",
    "ReverseDependenciesAutopkgtestWorkflow",
    "SbuildWorkflow",
    "UpdateEnvironmentsWorkflow",
    "UpdateSuiteWorkflow",
    "UpdateSuitesWorkflow",
    "Workflow",
    "WorkflowValidationError",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for TaskHistoryManager."""

import datetime as dt
from typing import override

from django.db.models import IntegerField, Q
from django.db.models.functions import Cast
from django.db.models.lookups import In
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebusineHistoricalTaskRun,
    RuntimeStatistics,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Collection, CollectionItem, WorkRequest
from debusine.db.models.lookups import LookupResultBare
from debusine.db.playground import scenarios
from debusine.server.collections.base import ItemAdditionError
from debusine.server.collections.lookup import lookup_multiple
from debusine.server.collections.task_history import TaskHistoryManager
from debusine.tasks.models import LookupMultiple
from debusine.test.django import TestCase


class TaskHistoryManagerTests(TestCase):
    """Tests for TaskHistoryManager."""

    scenario = scenarios.DefaultContext()
    workflow: WorkRequest
    collection: Collection
    collection_lookup: str
    manager: TaskHistoryManager

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.workflow = cls.playground.create_workflow()
        cls.collection = cls.scenario.workspace.get_singleton_collection(
            pc=cls.scenario.context_ui(),
            category=CollectionCategory.TASK_HISTORY,
        )
        cls.collection_lookup = f"_@{CollectionCategory.TASK_HISTORY}"
        cls.manager = TaskHistoryManager(collection=cls.collection)

    def create_historical_task_run(
        self,
        task_type: TaskTypes,
        task_name: str,
        subject: str | None,
        context: str | None,
        work_request_id: int,
        result: WorkRequest.Results = WorkRequest.Results.SUCCESS,
        collection: Collection | None = None,
    ) -> CollectionItem:
        """Create and return a historical task run in a collection."""
        created_at = dt.datetime(2020, 1, 1, tzinfo=dt.UTC) + dt.timedelta(
            seconds=work_request_id
        )
        return CollectionItem.objects.create_from_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            parent_collection=collection or self.collection,
            name=":".join(
                [
                    task_type,
                    task_name,
                    subject or "",
                    context or "",
                    str(work_request_id),
                ]
            ),
            data=DebusineHistoricalTaskRun(
                task_type=task_type,
                task_name=task_name,
                subject=subject,
                context=context,
                timestamp=int(created_at.timestamp()),
                work_request_id=work_request_id,
                result=result,
                runtime_statistics=RuntimeStatistics(duration=1),
            ),
            created_by_user=self.scenario.user,
            created_by_workflow=self.workflow,
        )

    def test_add_bare_data_no_data(self) -> None:
        """`add_bare_data` requires item data."""
        with self.assertRaisesRegex(
            ItemAdditionError, "Adding to debusine:task-history requires data"
        ):
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
            )

    def test_add_bare_data_refuses_name(self) -> None:
        """`add_bare_data` refuses an explicitly-specified name."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Cannot use an explicit item name when adding to "
            "debusine:task-history",
        ):
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=DebusineHistoricalTaskRun(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid/amd64",
                    timestamp=int(timezone.now().timestamp()),
                    work_request_id=1,
                    result=WorkRequest.Results.SUCCESS,
                    runtime_statistics=RuntimeStatistics(duration=1),
                ),
                name="override",
            )

    def test_add_bare_data_raise_item_addition_error(self) -> None:
        """`add_bare_data` raises an error for duplicate names."""
        data = DebusineHistoricalTaskRun(
            task_type=TaskTypes.WORKER,
            task_name="sbuild",
            subject="base-files",
            context="sid/amd64",
            timestamp=int(timezone.now().timestamp()),
            work_request_id=1,
            result=WorkRequest.Results.SUCCESS,
            runtime_statistics=RuntimeStatistics(duration=1),
        )
        self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=data,
        )

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=data,
            )

    def test_add_bare_data_different_work_request_ids(self) -> None:
        """Adding bare data items with different work request IDs is OK."""
        items = [
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=DebusineHistoricalTaskRun(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid/amd64",
                    timestamp=int(timezone.now().timestamp()),
                    work_request_id=work_request_id,
                    result=WorkRequest.Results.SUCCESS,
                    runtime_statistics=RuntimeStatistics(duration=1),
                ),
            )
            for work_request_id in (1, 2)
        ]

        self.assertEqual(
            [item.name for item in items],
            [
                "Worker:sbuild:base-files:sid/amd64:1",
                "Worker:sbuild:base-files:sid/amd64:2",
            ],
        )

    def test_add_bare_data_unset_subject_and_context(self) -> None:
        """`add_bare_data` handles subject=None and context=None."""
        item = self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=DebusineHistoricalTaskRun(
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                subject=None,
                context=None,
                timestamp=int(timezone.now().timestamp()),
                work_request_id=1,
                result=WorkRequest.Results.SUCCESS,
                runtime_statistics=RuntimeStatistics(duration=1),
            ),
        )

        self.assertEqual(item.name, "Worker:sbuild:::1")

    def test_add_bare_data_quotes_subject_and_context(self) -> None:
        """`add_bare_data` URL-encodes subject and context."""
        item = self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=DebusineHistoricalTaskRun(
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                subject="hello:world",
                context="any:amd64:amd64",
                timestamp=int(timezone.now().timestamp()),
                work_request_id=1,
                result=WorkRequest.Results.SUCCESS,
                runtime_statistics=RuntimeStatistics(duration=1),
            ),
        )

        self.assertEqual(
            item.name, "Worker:sbuild:hello%3Aworld:any%3Aamd64%3Aamd64:1"
        )

    def test_add_bare_data_replace(self) -> None:
        """`add_bare_data` can replace an existing bare data item."""
        data = DebusineHistoricalTaskRun(
            task_type=TaskTypes.WORKER,
            task_name="sbuild",
            subject="base-files",
            context="sid/amd64",
            timestamp=int(timezone.now().timestamp()),
            work_request_id=1,
            result=WorkRequest.Results.SUCCESS,
            runtime_statistics=RuntimeStatistics(duration=1),
        )
        item_old = self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=data,
        )

        item_new = self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )

        serialized_data = data.model_dump(mode="json", exclude_unset=True)
        item_old.refresh_from_db()
        self.assertEqual(item_old.name, "Worker:sbuild:base-files:sid/amd64:1")
        self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_old.data, serialized_data)
        self.assertEqual(item_old.removed_by_user, self.scenario.user)
        self.assertEqual(item_old.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(item_new.name, "Worker:sbuild:base-files:sid/amd64:1")
        self.assertEqual(item_new.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_new.data, serialized_data)
        self.assertIsNone(item_new.removed_at)

        # Adding another item doesn't touch item_old's removed_at, even if
        # we make the expiry policy more aggressive.
        self.collection.data = {"old_items_to_keep": 0}
        self.collection.save()
        previous_item_old_removed_at = item_old.removed_at
        self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )
        item_old.refresh_from_db()
        self.assertEqual(item_old.removed_at, previous_item_old_removed_at)

    def test_add_bare_data_replace_nonexistent(self) -> None:
        """Replacing a nonexistent bare data item is allowed."""
        data = DebusineHistoricalTaskRun(
            task_type=TaskTypes.WORKER,
            task_name="sbuild",
            subject="base-files",
            context="sid/amd64",
            timestamp=int(timezone.now().timestamp()),
            work_request_id=1,
            result=WorkRequest.Results.SUCCESS,
            runtime_statistics=RuntimeStatistics(duration=1),
        )

        item = self.manager.add_bare_data(
            BareDataCategory.HISTORICAL_TASK_RUN,
            user=self.scenario.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )

        self.assertEqual(item.name, "Worker:sbuild:base-files:sid/amd64:1")
        self.assertEqual(item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(
            item.data, data.model_dump(mode="json", exclude_unset=True)
        )

    def test_add_bare_data_keeps_last_success(self) -> None:
        """`add_bare_data` keeps the last successful run."""
        items: list[CollectionItem] = []
        for task_type, task_name, subject, context in (
            (TaskTypes.WORKER, "sbuild", "base-files", "sid/amd64"),
            (TaskTypes.WORKER, "sbuild", "base-files", "trixie/amd64"),
            (TaskTypes.WORKER, "sbuild", "hello", "sid/amd64"),
            (TaskTypes.SERVER, "aptmirror", "base-files", "sid/amd64"),
        ):
            for i in range(6):
                items.append(
                    self.create_historical_task_run(
                        task_type,
                        task_name,
                        subject,
                        context,
                        len(items) + 1,
                        result=(
                            WorkRequest.Results.SUCCESS
                            if i == 0
                            else WorkRequest.Results.FAILURE
                        ),
                    )
                )

        items.append(
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=DebusineHistoricalTaskRun(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid/amd64",
                    timestamp=int(timezone.now().timestamp()),
                    work_request_id=len(items) + 1,
                    result=WorkRequest.Results.FAILURE,
                    runtime_statistics=RuntimeStatistics(duration=1),
                ),
            )
        )

        # Matching items are pruned, but the most recent success is kept
        # despite not being within the window of old items to keep.
        self.assertQuerySetEqual(
            self.collection.child_items.active().order_by("id"),
            [items[0], *items[2:]],
        )

    def test_add_bare_data_keeps_last_failure(self) -> None:
        """`add_bare_data` keeps the last failed run."""
        items: list[CollectionItem] = []
        for task_type, task_name, subject, context in (
            (TaskTypes.WORKER, "sbuild", "base-files", "sid/amd64"),
            (TaskTypes.WORKER, "sbuild", "base-files", "trixie/amd64"),
            (TaskTypes.WORKER, "sbuild", "hello", "sid/amd64"),
            (TaskTypes.SERVER, "aptmirror", "base-files", "sid/amd64"),
        ):
            for i in range(6):
                items.append(
                    self.create_historical_task_run(
                        task_type,
                        task_name,
                        subject,
                        context,
                        len(items) + 1,
                        result=(
                            WorkRequest.Results.FAILURE
                            if i == 0
                            else WorkRequest.Results.SUCCESS
                        ),
                    )
                )

        items.append(
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=DebusineHistoricalTaskRun(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid/amd64",
                    timestamp=int(timezone.now().timestamp()),
                    work_request_id=len(items) + 1,
                    result=WorkRequest.Results.SUCCESS,
                    runtime_statistics=RuntimeStatistics(duration=1),
                ),
            )
        )

        # Matching items are pruned, but the most recent failure is kept
        # despite not being within the window of old items to keep.
        self.assertQuerySetEqual(
            self.collection.child_items.active().order_by("id"),
            [items[0], *items[2:]],
        )

    def test_add_bare_data_old_items_to_keep(self) -> None:
        """The number of old items to keep is configurable."""
        self.collection.data = {"old_items_to_keep": 3}
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "sbuild", "base-files", "sid/amd64", i + 1
            )
            for i in range(5)
        ]

        items.append(
            self.manager.add_bare_data(
                BareDataCategory.HISTORICAL_TASK_RUN,
                user=self.scenario.user,
                workflow=self.workflow,
                data=DebusineHistoricalTaskRun(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid/amd64",
                    timestamp=int(timezone.now().timestamp()),
                    work_request_id=len(items) + 1,
                    result=WorkRequest.Results.SUCCESS,
                    runtime_statistics=RuntimeStatistics(duration=1),
                ),
            )
        )

        self.assertQuerySetEqual(
            self.collection.child_items.active().order_by("id"), items[3:]
        )

    def test_remove_item_bare_data(self) -> None:
        """`remove_item` removes a bare data item."""
        item = self.create_historical_task_run(
            TaskTypes.WORKER, "sbuild", "base-files", "sid/amd64", 1
        )

        self.manager.remove_item(
            item, user=self.scenario.user, workflow=self.workflow
        )

        self.assertEqual(item.removed_by_user, self.scenario.user)
        self.assertEqual(item.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item.removed_at)

    def test_lookup_unexpected_format(self) -> None:
        """`lookup` raises `LookupError` for an unexpected format."""
        with self.assertRaisesRegex(
            LookupError,
            r'^Unexpected lookup format: "foo:bar" \(expected '
            r'"last-entry", "last-success", or "last-failure"\)$',
        ):
            self.manager.lookup("foo:bar")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(
            self.manager.lookup("last-entry:Worker:sbuild:base-files:sid/amd64")
        )
        self.assertIsNone(
            self.manager.lookup(
                "last-success:Worker:sbuild:base-files:sid/amd64"
            )
        )
        self.assertIsNone(
            self.manager.lookup(
                "last-failure:Worker:sbuild:base-files:sid/amd64"
            )
        )
        self.assertIsNone(self.manager.lookup("name:nonexistent"))

    def test_lookup_last_entry(self) -> None:
        """`lookup` for `last-entry` returns the last matching entry."""
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "sbuild", subject, context, work_request_id
            )
            for subject, context, work_request_id in (
                (None, None, 1),
                ("base-files", "sid/amd64", 2),
                ("base-passwd", "sid/amd64", 3),
                ("base-files", "sid/amd64", 4),
                ("hello:world", "any:amd64:amd64", 5),
            )
        ]

        self.assertEqual(
            self.manager.lookup("last-entry:Worker:sbuild::"), items[0]
        )
        self.assertEqual(
            self.manager.lookup(
                "last-entry:Worker:sbuild:base-files:sid/amd64"
            ),
            items[3],
        )
        self.assertEqual(
            self.manager.lookup(
                "last-entry:Worker:sbuild:base-passwd:sid/amd64"
            ),
            items[2],
        )
        self.assertEqual(
            self.manager.lookup(
                "last-entry:Worker:sbuild:hello%3Aworld:any%3Aamd64%3Aamd64"
            ),
            items[4],
        )

    def test_lookup_last_success(self) -> None:
        """`lookup` for `last-success` returns the last matching success."""
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER,
                "sbuild",
                subject,
                context,
                work_request_id,
                result=result,
            )
            for subject, context, work_request_id, result in (
                (None, None, 1, WorkRequest.Results.SUCCESS),
                ("base-files", "sid/amd64", 2, WorkRequest.Results.SUCCESS),
                ("base-passwd", "sid/amd64", 3, WorkRequest.Results.SUCCESS),
                ("base-files", "sid/amd64", 4, WorkRequest.Results.FAILURE),
                (
                    "hello:world",
                    "any:amd64:amd64",
                    5,
                    WorkRequest.Results.SUCCESS,
                ),
            )
        ]

        self.assertEqual(
            self.manager.lookup("last-success:Worker:sbuild::"), items[0]
        )
        self.assertEqual(
            self.manager.lookup(
                "last-success:Worker:sbuild:base-files:sid/amd64"
            ),
            items[1],
        )
        self.assertEqual(
            self.manager.lookup(
                "last-success:Worker:sbuild:hello%3Aworld:any%3Aamd64%3Aamd64"
            ),
            items[4],
        )

    def test_lookup_last_failure(self) -> None:
        """`lookup` for `last-failure` returns the last matching failure."""
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER,
                "sbuild",
                subject,
                context,
                work_request_id,
                result=result,
            )
            for subject, context, work_request_id, result in (
                (None, None, 1, WorkRequest.Results.FAILURE),
                ("base-files", "sid/amd64", 2, WorkRequest.Results.FAILURE),
                ("base-passwd", "sid/amd64", 3, WorkRequest.Results.FAILURE),
                ("base-files", "sid/amd64", 4, WorkRequest.Results.SUCCESS),
                (
                    "hello:world",
                    "any:amd64:amd64",
                    5,
                    WorkRequest.Results.FAILURE,
                ),
            )
        ]

        self.assertEqual(
            self.manager.lookup("last-failure:Worker:sbuild::"), items[0]
        )
        self.assertEqual(
            self.manager.lookup(
                "last-failure:Worker:sbuild:base-files:sid/amd64"
            ),
            items[1],
        )
        self.assertEqual(
            self.manager.lookup(
                "last-failure:Worker:sbuild:hello%3Aworld:any%3Aamd64%3Aamd64"
            ),
            items[4],
        )

    def test_lookup_filter_unexpected_format(self) -> None:
        """`lookup_filter` raises LookupError for an unexpected format."""
        with self.assertRaisesRegex(
            LookupError,
            r'^Unexpected lookup filter format: "invalid" \(expected '
            r'"same_work_request" or "same_workflow"\)$',
        ):
            self.manager.lookup_filter(
                "invalid",
                "foo",
                workspace=self.scenario.workspace,
                pc=self.scenario.context_ui(),
            )

    def test_lookup_filter_same_work_request_single(self) -> None:
        """`lookup_filter`: `same_work_request` with a single lookup."""
        work_requests = [self.playground.create_worker_task() for _ in range(2)]
        upload, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD, work_request=work_requests[0]
        )
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "noop", "subject", "context", work_request.id
            )
            for work_request in work_requests
        ]
        subordinate_lookup = f"{upload.id}@artifacts"

        condition = self.manager.lookup_filter(
            "same_work_request",
            subordinate_lookup,
            workspace=self.scenario.workspace,
            pc=self.scenario.context_ui(),
        )
        self.assertEqual(
            condition,
            Q(
                In(
                    Cast("data__work_request_id", IntegerField()),
                    {work_requests[0].id},
                )
            ),
        )
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition), [items[0]]
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "child_type": LookupChildType.BARE,
                "lookup__same_work_request": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.scenario.workspace,
                pc=self.scenario.context_ui(),
                expect_type=LookupChildType.BARE,
            ),
            [
                LookupResultBare(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=items[0],
                )
            ],
        )

    def test_lookup_filter_same_work_request_multiple(self) -> None:
        """`lookup_filter`: `same_work_request` with a multiple lookup."""
        work_requests = [self.playground.create_worker_task() for _ in range(3)]
        uploads = [
            self.playground.create_artifact(
                category=ArtifactCategory.UPLOAD, work_request=work_request
            )[0]
            for work_request in work_requests[:2]
        ]
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "noop", "subject", "context", work_request.id
            )
            for work_request in work_requests
        ]
        subordinate_lookup = LookupMultiple.model_validate(
            [f"{upload.id}@artifacts" for upload in uploads]
        )

        condition = self.manager.lookup_filter(
            "same_work_request",
            subordinate_lookup,
            workspace=self.scenario.workspace,
            pc=self.scenario.context_ui(),
        )
        self.assertEqual(
            condition,
            Q(
                In(
                    Cast("data__work_request_id", IntegerField()),
                    {work_request.id for work_request in work_requests[:2]},
                )
            ),
        )
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition),
            items[:2],
            ordered=False,
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "child_type": LookupChildType.BARE,
                "lookup__same_work_request": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.scenario.workspace,
                pc=self.scenario.context_ui(),
                expect_type=LookupChildType.BARE,
            ),
            [
                LookupResultBare(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=item,
                )
                for item in items[:2]
            ],
        )

    def test_lookup_filter_same_workflow_single(self) -> None:
        """`lookup_filter`: `same_workflow` with a single lookup."""
        work_requests = [
            self.workflow.create_child_worker("noop") for _ in range(2)
        ]
        work_requests.append(self.playground.create_worker_task())
        upload, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD, work_request=work_requests[0]
        )
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "noop", "subject", "context", work_request.id
            )
            for work_request in work_requests
        ]
        subordinate_lookup = f"{upload.id}@artifacts"

        condition = self.manager.lookup_filter(
            "same_workflow",
            subordinate_lookup,
            workspace=self.scenario.workspace,
            pc=self.scenario.context_ui(),
        )
        # Ideally we'd check the query itself, but digging into the
        # WorkRequest subquery is hard, so we just check the results.
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition),
            items[:2],
            ordered=False,
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "child_type": LookupChildType.BARE,
                "lookup__same_workflow": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.scenario.workspace,
                pc=self.scenario.context_ui(),
                expect_type=LookupChildType.BARE,
            ),
            [
                LookupResultBare(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=item,
                )
                for item in items[:2]
            ],
        )

    def test_lookup_filter_same_workflow_multiple(self) -> None:
        """`lookup_filter`: `same_workflow` with a multiple lookup."""
        work_requests = [
            self.workflow.create_child_worker("noop") for _ in range(2)
        ]
        other_workflow = self.playground.create_workflow()
        work_requests.extend(
            [other_workflow.create_child_worker("noop") for _ in range(2)]
        )
        work_requests.append(self.playground.create_worker_task())
        uploads = [
            self.playground.create_artifact(
                category=ArtifactCategory.UPLOAD, work_request=work_request
            )[0]
            for work_request in (work_requests[0], work_requests[2])
        ]
        items = [
            self.create_historical_task_run(
                TaskTypes.WORKER, "noop", "subject", "context", work_request.id
            )
            for work_request in work_requests
        ]
        subordinate_lookup = LookupMultiple.model_validate(
            [f"{upload.id}@artifacts" for upload in uploads]
        )

        condition = self.manager.lookup_filter(
            "same_workflow",
            subordinate_lookup,
            workspace=self.scenario.workspace,
            pc=self.scenario.context_ui(),
        )
        # Ideally we'd check the query itself, but digging into the
        # WorkRequest subquery is hard, so we just check the results.
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition),
            items[:4],
            ordered=False,
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "child_type": LookupChildType.BARE,
                "lookup__same_workflow": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.scenario.workspace,
                pc=self.scenario.context_ui(),
                expect_type=LookupChildType.BARE,
            ),
            [
                LookupResultBare(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=item,
                )
                for item in items[:4]
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Lintian class."""

import datetime as dt
import itertools
import json
import textwrap
from pathlib import Path
from subprocess import CompletedProcess
from typing import Any, cast, override
from unittest import mock
from unittest.mock import MagicMock, call

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import LintianArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianLintianSummary,
    DebianSourcePackage,
    DebianUpload,
    WorkRequestResults,
)
from debusine.client.models import FileResponse, FilesResponseType
from debusine.tasks import Lintian, TaskConfigError
from debusine.tasks.executors import InstanceInterface
from debusine.tasks.models import (
    BackendType,
    BaseTaskInputArtifact,
    LintianDynamicData,
    LintianFailOnSeverity,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
    create_system_tarball_data,
)


class LintianTests(ExternalTaskHelperMixin[Lintian], TestCase):
    """Tests for Lintian."""

    SAMPLE_TASK_DATA = {
        "input": {
            "source_artifact": 5,
        },
        "environment": "debian/match:codename=bookworm:architecture=amd64",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name="hello",
                version="1.0",
                type="dpkg",
                dsc_fields={},
            ),
        )
        self.resolver.set_field_artifacts("input_binary_artifacts")

        # Different tests might try to get the lintian version when
        # preparing the summary. It's not relevant and the CI system
        # might not have "lintian" installed
        self._get_lintian_version_patcher = self.patch_get_lintian_version(
            "1.0.0"
        )

    def assert_parse_output(
        self, output: str, expected: list[dict[str, str]]
    ) -> None:
        """
        Assert parsed output from Lintian._parse_output is as expected.

        :param output: A string that is saved into a file to be used
          by Lintian._parse_output
        :param expected: A list containing the expected parsed output
        """
        self.assertEqual(
            Lintian._parse_output(
                self.create_temporary_file(contents=output.encode("utf-8"))
            ),
            expected,
        )

    def test_parse_output_empty(self) -> None:
        """Parse an empty output."""
        output = ""
        expected: list[dict[str, str]] = []

        self.assert_parse_output(output, expected)

    def test_parse_output_only_N_lines(self) -> None:
        """Parse an output with only N: lines. They are ignored."""
        output = (
            "N:   This package does not use a machine-readable...\n"
            "N:   second line\n"
        )
        expected: list[dict[str, str]] = []

        self.assert_parse_output(output, expected)

    def test_parse_output_simple(self) -> None:
        """Parse tag without note, neither pointer."""
        output = "E: qnetload: copyright-contains-dh_make-todo-boilerplate"
        expected = [
            {
                "severity": "error",
                "package": "qnetload",
                "tag": "copyright-contains-dh_make-todo-boilerplate",
                "note": "",
                "pointer": "",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_tag_line_bullseye(self) -> None:
        """Parse tag line on bullseye that is duplicated information: ignore."""
        output = "W: package-uses-deprecated-debhelper-compat-version"
        self.assertIsNone(Lintian._parse_tag_line(output))

    def test_parse_output_bullseye(self) -> None:
        """Ignore line with "W: {tag}" (not including package name)."""
        output = textwrap.dedent(
            """\
            W: hello source: package-uses-deprecated-debhelper-compat-version 9
            N:
            W: package-uses-deprecated-debhelper-compat-version
            N:
            N: The debhelper compatibility version used by this package...
            N:
            N: The compatibility version can be set...
            N:
            W: hello source: national-encoding debian/po/sv.po
            N:
            W: national-encoding
            N:
            N: A file is not valid UTF-8.
            N:
            N: Debian has used UTF-8 for many years...
            """
        )
        explanation_1 = textwrap.dedent(
            """\
            The debhelper compatibility version used by this package...

            The compatibility version can be set..."""
        )
        explanation_2 = textwrap.dedent(
            """\
            A file is not valid UTF-8.

            Debian has used UTF-8 for many years..."""
        )

        expected = [
            {
                "comment": "",
                "explanation": explanation_1,
                "note": "9",
                "package": "hello source",
                "pointer": "",
                "severity": "warning",
                "tag": "package-uses-deprecated-debhelper-compat-version",
            },
            {
                "comment": "",
                "explanation": explanation_2,
                "note": "debian/po/sv.po",
                "package": "hello source",
                "pointer": "",
                "severity": "warning",
                "tag": "national-encoding",
            },
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_with_explanation(self) -> None:
        """Parse tag with explanation."""
        output = textwrap.dedent(
            """\
            N:
            I: hello: hardening-no-bindnow [usr/bin/hello]
            N:
            N:   This package provides an ELF...
            N:
            N:   This is needed (together with "relro")...
            N:
            N:
            I: hello: typo-in-manual-page add [usr/share/man/man1/hello.1.gz:27]
            N:
            N:   Test
            """
        )

        explanation_1 = textwrap.dedent(
            """\
            This package provides an ELF...

            This is needed (together with "relro")..."""
        )

        explanation_2 = "Test"

        expected = [
            {
                "severity": "info",
                "package": "hello",
                "tag": "hardening-no-bindnow",
                "note": "",
                "pointer": "usr/bin/hello",
                "explanation": explanation_1,
                "comment": "",
            },
            {
                "severity": "info",
                "package": "hello",
                "tag": "typo-in-manual-page",
                "note": "add",
                "pointer": "usr/share/man/man1/hello.1.gz:27",
                "explanation": explanation_2,
                "comment": "",
            },
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_overridden(self) -> None:
        """Parse output with overridden and a comment."""
        output = textwrap.dedent(
            """\
            N:
            I: hello source: hardening-no-bindnow [usr/bin/hello]
            N:
            N:   This package provides an ELF...
            N:
            N:   This is needed (together with "relro")...
            N:
            N:
            N: Lintian detect a source file but is a hand generated example so
            N: it is ignored
            O: python-cloudscraper source: source-is-missing [tests/fixtures/js_challenge-27-05-2020.html]
            N:
            N:   The source of the following file is missing. Lintian checked a few
            N:   possible paths to find the source, and did not find it.
            N:
            N:   Please repack your package to include the source or add it to
            N:   "debian/missing-sources" directory.
            N:
            N:   Please note, that very-long-line-length-in-source-file tagged
            N:   files are likely tagged source-is-missing. It is a feature
            N:   not a bug.
            N:
            N:   Visibility: error
            N:   Show-Always: no
            N:   Check: files/source-missing
            N:
            N:
            N: Same as before
            O: python-cloudscraper source: source-is-missing [tests/fixtures/js_challenge1_16_05_2020.html]
            N:
            N:
            """  # noqa: E501
        )

        explanation_info = textwrap.dedent(
            """\
            This package provides an ELF...

            This is needed (together with "relro")..."""
        )

        explanation_overridden_1 = textwrap.dedent(
            """\
            The source of the following file is missing. Lintian checked a few
            possible paths to find the source, and did not find it.

            Please repack your package to include the source or add it to
            "debian/missing-sources" directory.

            Please note, that very-long-line-length-in-source-file tagged
            files are likely tagged source-is-missing. It is a feature
            not a bug.

            Visibility: error
            Show-Always: no
            Check: files/source-missing"""
        )

        comment_overridden_1 = textwrap.dedent(
            """\
            Lintian detect a source file but is a hand generated example so
            it is ignored"""
        )

        explanation_overridden_2 = ""
        comment_overridden_2 = "Same as before"

        expected = [
            {
                "severity": "info",
                "package": "hello source",
                "tag": "hardening-no-bindnow",
                "note": "",
                "pointer": "usr/bin/hello",
                "explanation": explanation_info,
                "comment": "",
            },
            {
                "severity": "overridden",
                "package": "python-cloudscraper source",
                "tag": "source-is-missing",
                "note": "",
                "pointer": "tests/fixtures/js_challenge-27-05-2020.html",
                "explanation": explanation_overridden_1,
                "comment": comment_overridden_1,
            },
            {
                "severity": "overridden",
                "package": "python-cloudscraper source",
                "tag": "source-is-missing",
                "note": "",
                "pointer": "tests/fixtures/js_challenge1_16_05_2020.html",
                "explanation": explanation_overridden_2,
                "comment": comment_overridden_2,
            },
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_source(self) -> None:
        """Parse tag with "package_name source."."""
        output = (
            "P: hello source: license-problem-gfdl-non-official-text invariant "
            "part is: with no invariant"
        )
        expected = [
            {
                "severity": "pedantic",
                "package": "hello source",
                "tag": "license-problem-gfdl-non-official-text",
                "note": "invariant part is: with no invariant",
                "pointer": "",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_with_information_one_word(self) -> None:
        """Parse tag with note."""
        output = (
            "E: qnetload: description-contains-invalid-control-statement line 3"
        )
        expected = [
            {
                "severity": "error",
                "package": "qnetload",
                "tag": "description-contains-invalid-control-statement",
                "note": "line 3",
                "pointer": "",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_with_information_and_file(self) -> None:
        """Parse tag for a source package with note and pointer."""
        output = (
            "E: qnetload source: the-tag some information [debian/changelog:33]"
        )
        expected = [
            {
                "severity": "error",
                "package": "qnetload source",
                "tag": "the-tag",
                "note": "some information",
                "pointer": "debian/changelog:33",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_with_file(self) -> None:
        """Parse tag with pointer."""
        output = (
            "E: qnetload: changelog-is-dh_make-template "
            "[usr/share/doc/qnetload/changelog.Debian.gz:1]"
        )
        expected = [
            {
                "severity": "error",
                "package": "qnetload",
                "tag": "changelog-is-dh_make-template",
                "note": "",
                "pointer": "usr/share/doc/qnetload/changelog.Debian.gz:1",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_severity_masked_ignored(self) -> None:
        """Parse output severity is Masked: ignored."""
        output = (
            "M: xserver-xorg-video-sisusb source: "
            "very-long-line-length-in-source-file 705 > 512 [configure:15910]"
        )
        expected: list[dict[str, str]] = []

        self.assert_parse_output(output, expected)

    def test_parse_output_severity_c(self) -> None:
        """Parse output severity is Classification."""
        output = (
            "C: xserver-xorg-video-sisusb source: "
            "very-long-line-length-in-source-file 705 > 512 [configure:15910]"
        )
        expected = [
            {
                "severity": "classification",
                "tag": "very-long-line-length-in-source-file",
                "package": "xserver-xorg-video-sisusb source",
                "note": "705 > 512",
                "pointer": "configure:15910",
                "explanation": "",
                "comment": "",
            }
        ]

        self.assert_parse_output(output, expected)

    def test_parse_output_raise_value_error(self) -> None:
        """Output cannot be parsed."""
        line = "XXX"
        with self.assertRaisesRegex(
            ValueError, f"Failed to parse line: {line}"
        ):
            Lintian._parse_output(
                self.create_temporary_file(contents=line.encode("utf-8"))
            )

    def test_task_result_empty_file_success(self) -> None:
        """task_result() for an empty file return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / Lintian.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task(
            override={"fail_on_severity": LintianFailOnSeverity.WARNING}
        )
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_severity_less_than_fail_on_success(self) -> None:
        """task_result() severity < than fail_on_severity: return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / Lintian.CAPTURE_OUTPUT_FILENAME).write_text(
            "W: cynthiune.app: vcs-obsolete"
        )

        self.task._package_name_to_filename = {
            "cynthiune.app": "cynthiune-app.deb"
        }
        self.task._architecture_to_packages["all"] = {"cynthiune.app"}

        self.configure_task(
            override={"fail_on_severity": LintianFailOnSeverity.ERROR}
        )

        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_severity_equal_as_fail_on_failure(self) -> None:
        """task_result() severity == fail_on_severity: return FAILURE."""
        directory = self.create_temporary_directory()
        (directory / Lintian.CAPTURE_OUTPUT_FILENAME).write_text(
            "W: cynthiune.app: vcs-obsolete"
        )

        self.configure_task(
            override={"fail_on_severity": LintianFailOnSeverity.WARNING}
        )

        self.task._package_name_to_filename = {
            "cynthiune.app": "cynthiune-app.deb"
        }
        self.task._architecture_to_packages["all"] = {"cynthiune.app"}

        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.FAILURE
        )

    def test_task_result_fail_on_none(self) -> None:
        """task_result() fail_on_severity == "none": return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / Lintian.CAPTURE_OUTPUT_FILENAME).write_text(
            "E: cynthiune.app: vcs-obsolete\nW: hello: invalid-something\n"
        )

        self.configure_task(
            override={"fail_on_severity": LintianFailOnSeverity.NONE}
        )

        self.task._package_name_to_filename = {
            "cynthiune.app": "cynthiune-app.deb",
            "hello": "hello.deb",
        }
        self.task._architecture_to_packages["all"] = {"cynthiune.app"}
        self.task._architecture_to_packages["amd64"] = {"hello"}

        # Succeeded (fail_on_severity: none)
        self.assertEqual(
            self.task.task_result(2, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_runtime_error(self) -> None:
        directory = self.create_temporary_directory()
        (directory / Lintian.CAPTURE_OUTPUT_FILENAME).touch()

        self.assertEqual(
            self.task.task_result(1, directory), WorkRequestResults.ERROR
        )

    @staticmethod
    def get_analysis_no_tags() -> dict[str, Any]:
        """Return a basic analysis without any Lintian tags."""
        return {
            "summary": {
                "distribution": "debian:unstable",
                "lintian_version": "1.0.0",
                "overridden_tags_found": [],
                "package_filename": {},
                "tags_count_by_severity": {
                    "classification": 0,
                    "error": 0,
                    "experimental": 0,
                    "info": 0,
                    "overridden": 0,
                    "pedantic": 0,
                    "warning": 0,
                },
                "tags_found": [],
            },
            "tags": [],
            "version": 1,
        }

    def assert_architecture_to_analysis(
        self,
        output_config: dict[str, bool],
        expected_architecture_to_analysis: dict[str, dict[str, Any]],
    ) -> None:
        """
        _create_architecture_to_analysis return the expected output.

        The real analysis test is implemented in test_create_analysis(). This
        test is only to check that only the relevant architectures are
        returned.
        """
        self.configure_task(override={"output": output_config})
        for architecture in ("source", "all", "amd64"):
            self.task._architecture_to_packages[architecture] = set()

        lintian_file = self.create_temporary_file()
        lintian_file.write_text(
            "E: cynthiune.app: vcs-obsolete\n"
            "W: hello source: invalid-something\n"
        )

        self.assertEqual(
            self.task._create_architecture_to_analysis(lintian_file),
            expected_architecture_to_analysis,
        )

    def test_analyze_lintian_output_no_output_tasks(self) -> None:
        """analyze_lintian_output(): no analysis returned."""
        analysis_config = {
            "source_analysis": False,
            "binary_all_analysis": False,
            "binary_any_analysis": False,
        }
        expected_analysis: dict[str, dict[str, Any]] = {}
        self.assert_architecture_to_analysis(analysis_config, expected_analysis)

    def test_create_architecture_to_analysis_only_source(self) -> None:
        """test_create_architecture_to_analysis(): only source analysed."""
        analysis_config = {
            "source_analysis": True,
            "binary_all_analysis": False,
            "binary_any_analysis": False,
        }
        expected_analysis = {"source": self.get_analysis_no_tags()}
        self.assert_architecture_to_analysis(analysis_config, expected_analysis)

    def test_create_architecture_to_analysis_only_binary_all(self) -> None:
        """test_create_architecture_to_analysis(): only binary_all analysed."""
        analysis_config = {
            "source_analysis": False,
            "binary_all_analysis": True,
            "binary_any_analysis": False,
        }
        expected_analysis = {"all": self.get_analysis_no_tags()}
        self.assert_architecture_to_analysis(analysis_config, expected_analysis)

    def test_create_architecture_to_analysis_only_source_and_binary_all(
        self,
    ) -> None:
        """analyze_lintian_output(): only source and binary all analysed."""
        analysis_config = {
            "source_analysis": True,
            "binary_all_analysis": True,
            "binary_any_analysis": False,
        }
        expected_analysis = {
            "all": self.get_analysis_no_tags(),
            "source": self.get_analysis_no_tags(),
        }
        self.assert_architecture_to_analysis(analysis_config, expected_analysis)

    def test_create_architecture_to_analysis_only_binary_any(self) -> None:
        """analyze_lintian_output(): only binary any analysed."""
        analysis_config = {
            "source_analysis": False,
            "binary_all_analysis": False,
            "binary_any_analysis": True,
        }
        expected_analysis = {"amd64": self.get_analysis_no_tags()}
        self.assert_architecture_to_analysis(analysis_config, expected_analysis)

    def test_configure(self) -> None:
        """configure() with valid data. No exception is raised."""
        self.configure_task()

    def test_configure_invalid_data(self) -> None:
        """configure() with invalid data used, TaskConfigError is raised."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"extra_field": "something"})

    def test_configure_with_target_distribution(self) -> None:
        """Configuration included "target_distribution". Saved no exceptions."""
        distribution = "bookworm"
        self.configure_task(override={"target_distribution": distribution})
        self.assertEqual(self.task.data.target_distribution, distribution)

    def test_configure_with_fail_on_severity(self) -> None:
        """Configuration included "fail_on_severity". Saved, no exception."""
        severity = LintianFailOnSeverity.WARNING
        self.configure_task(override={"fail_on_severity": severity})
        self.assertEqual(self.task.data.fail_on_severity, severity)

    def test_configure_without_fail_on_severity(self) -> None:
        """Configuration does not include "fail_on_severity". Default is set."""
        self.configure_task()
        self.assertEqual(
            self.task.data.fail_on_severity, LintianFailOnSeverity.ERROR
        )
        self.assertEqual(self.task.data.target_distribution, "debian:unstable")

    def test_configure_without_output(self) -> None:
        """Configuration does not include "output". Check defaults."""
        self.configure_task()
        self.assertEqual(
            self.task.data.output.model_dump(),
            {
                "source_analysis": True,
                "binary_all_analysis": True,
                "binary_any_analysis": True,
            },
        )

    def test_configure_without_backend(self) -> None:
        """Configuration does not include "backend". Check default is "auto"."""
        self.configure_task()
        self.assertEqual(self.task.data.backend, BackendType.AUTO)

    def test_configure_include_backend_unshare(self) -> None:
        """Configuration with backend unshare is valid."""
        self.configure_task(override={"backend": BackendType.UNSHARE})
        self.assertEqual(self.task.data.backend, BackendType.UNSHARE)

    def test_configure_without_any_input(self) -> None:
        """Configuration invalid: needs a source."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {}})

    def test_configure_valid_task_artifact_id(self) -> None:
        """Configuration valid: does not raise any exception."""
        self.configure_task(override={"input": {"source_artifact": 5}})

    def test_configure_valid_task_binary_ids(self) -> None:
        """Configure valid: does not raise any exception."""
        self.configure_task(override={"input": {"binary_artifacts": [6, 7]}})

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:lintian:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            },
        )

    def test_compute_scheduler_tags_no_source_artifact(self) -> None:
        self.resolver.field_values["input_source_artifact"] = None
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=6,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="hello-bin",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "hello-bin",
                        "Version": "1.0",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                ),
            ),
        )

        provided, required = self.compute_scheduler_tags()
        self.assertEqual(provided.sets, {})
        self.assertEqual(
            required.sets,
            {
                wtags.ProvenanceRequired.SYSTEM: {
                    wtags.BUILD_ARCH_PREFIX + "amd64",
                    wtags.EXECUTOR_PREFIX + "unshare",
                }
            },
        )

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            LintianDynamicData(
                environment_id=1,
                input_source_artifact_id=5,
                subject="hello",
                parameter_summary="hello_1.0",
                runtime_context="binary-all+binary-any+source:sid",
                configuration_context="sid",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_subject_no_inputs(self) -> None:
        self.resolver.field_values["input_source_artifact"] = None
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "sid")

    def test_compute_dynamic_data_subject_debian_upload(self) -> None:
        """compute_dynamic_data subject is package name of source artifact."""
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Source": "hello",
                    "Version": "1.0",
                    "Architecture": "amd64",
                    "Files": [{"name": "hello_1.0_amd64.deb"}],
                },
            ),
        )

        self.assertEqual(self.compute_dynamic_data(self.task).subject, "hello")

    def test_compute_dynamic_data_subject_runtime_context(self) -> None:
        """compute_dynamic_data runtime_context source only."""
        self.configure_task(
            override={
                "output": {
                    "source_analysis": True,
                    "binary_all_analysis": False,
                    "binary_any_analysis": False,
                }
            }
        )
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Source": "hello",
                    "Version": "1.0",
                    "Architecture": "amd64",
                    "Files": [{"name": "hello_1.0_amd64.deb"}],
                },
            ),
        )

        dynamic_data = self.compute_dynamic_data(self.task)

        self.assertEqual(dynamic_data.subject, "hello")
        self.assertEqual(dynamic_data.runtime_context, "source:sid")

    def test_compute_dynamic_data_debian_binary_package(self) -> None:
        """
        compute_dynamic_data subject is package name of source artifact.

        There is a binary artifact with other package names.
        """
        self.configure_task(
            override={
                "input": {"source_artifact": 5, "binary_artifacts": [6]},
                "output": {"source_analysis": False},
            }
        )
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Source": "hello",
                    "Architecture": "amd64",
                    "Version": "1.0",
                    "Files": [{"name": "hello_1.0_amd64.deb"}],
                },
            ),
        )
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=6,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="hello-bin",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "hello-bin",
                        "Version": "1.0",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                ),
            ),
        )

        dynamic_data = self.compute_dynamic_data(self.task)
        self.assertEqual(dynamic_data.subject, "hello")
        self.assertEqual(dynamic_data.parameter_summary, "hello_1.0")
        self.assertEqual(
            dynamic_data.runtime_context, "binary-all+binary-any:sid"
        )

    def test_compute_dynamic_data_source_artifact_subject_no_binaries(
        self,
    ) -> None:
        """compute_dynamic_data set to source artifact without binaries."""
        self.configure_task(remove=["environment"])
        self.resolver.field_values["environment"] = None
        self.resolver.set_field_artifact(
            "input_source_artifact",
            id=5,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Source": "hello",
                    "Version": "1.0",
                    "Architecture": "amd64",
                    "Files": [{"name": "hello_1.0_amd64.deb"}],
                },
            ),
        )

        dynamic_data = self.compute_dynamic_data(self.task)
        self.assertEqual(dynamic_data.subject, "hello")
        self.assertIsNone(dynamic_data.runtime_context)

    def test_compute_dynamic_data_debian_binary_package_subject(self) -> None:
        """
        compute_dynamic_data set subject to binary package name.

        There are no source packages.
        """
        self.configure_task(override={"input": {"binary_artifacts": [6]}})
        self.resolver.field_values["input_source_artifact"] = None
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=6,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="hello",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "hello",
                        "Version": "1.0",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                ),
            ),
        )

        self.assertEqual(self.compute_dynamic_data(self.task).subject, "hello")

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            Lintian.inputs["input_source_artifact"].categories,  # type: ignore[attr-defined]
            (
                ArtifactCategory.SOURCE_PACKAGE,
                ArtifactCategory.UPLOAD,
            ),
        )
        self.assertEqual(
            Lintian.inputs["input_binary_artifacts"].categories,  # type: ignore[attr-defined]
            (
                ArtifactCategory.BINARY_PACKAGE,
                ArtifactCategory.UPLOAD,
            ),
        )
        self.assertEqual(
            Lintian.inputs["environment"].categories,  # type: ignore[attr-defined]
            (
                ArtifactCategory.SYSTEM_TARBALL,
                ArtifactCategory.SYSTEM_IMAGE,
            ),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = LintianDynamicData(
            input_binary_artifacts_ids=[1, 2],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2])

        self.task.dynamic_data = LintianDynamicData(
            input_source_artifact_id=1,
            input_binary_artifacts_ids=[2, 3],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3])

        self.task.dynamic_data = LintianDynamicData(
            input_source_artifact_id=1,
            environment_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4])

    def test_cmdline_minimum_options(self) -> None:
        """Test Lintian._cmdline minimum options."""
        minimum_cmd = [
            "lintian",
            "--no-cfg",
            "--display-level",
            ">=classification",
            "--display-experimental",
            "--info",
        ]
        self.configure_task()
        self.task._lintian_targets = [self.create_temporary_file()]
        self.assertEqual(self.task._cmdline()[: len(minimum_cmd)], minimum_cmd)

    def test_cmdline_with_tags(self) -> None:
        """Cmdline add --tags."""
        tags = ["wish-script-but-no-wish-dep", "zero-byte-executable-in-path"]
        self.configure_task(override={"include_tags": tags})
        self.task._lintian_targets = [self.create_temporary_file()]

        cmdline = self.task._cmdline()

        self.assertIn(f"--tags={','.join(tags)}", cmdline)
        self.assertEqual(cmdline[-1], str(self.task._lintian_targets[0]))

    def test_cmdline_exclude_tags(self) -> None:
        """Cmdline add --suppress-tags."""
        tags = ["wayward-symbolic-link-target-in-source", "wrong-team"]

        self.configure_task(override={"exclude_tags": tags})
        self.task._lintian_targets = [self.create_temporary_file()]
        self.assertIn(f"--suppress-tags={','.join(tags)}", self.task._cmdline())

    def test_configure_for_execution_set_lintian_target(self) -> None:
        """One .dsc file: self.task._lintian_targets set to it."""
        directory = self.create_temporary_directory()

        dsc = self.create_temporary_file(suffix=".dsc", directory=directory)
        self.create_temporary_file(suffix=".tar.xz", directory=directory)

        self.patch_extract_package_name_arch().return_value = None, None

        prepare_executor_instance_mocked = (
            self.patch_prepare_executor_instance()
        )

        self.assertTrue(self.task.configure_for_execution(directory))

        prepare_executor_instance_mocked.assert_called_with()

        cast(MagicMock, self.task.executor_instance).assert_has_calls(
            [
                call.run(
                    ["apt-get", "update"],
                    run_as_root=True,
                    check=True,
                    stdout=mock.ANY,
                    stderr=mock.ANY,
                ),
                call.run(
                    ["apt-get", "--yes", "install", "lintian"],
                    run_as_root=True,
                    check=True,
                    stdout=mock.ANY,
                    stderr=mock.ANY,
                ),
            ]
        )

        self.assertEqual(self.task._lintian_targets, [dsc])

    def test_configure_no_dsc_file(self) -> None:
        """No files: configure_for_execution() return False."""
        download_directory = self.create_temporary_directory()

        ignored_file = "test.txt"

        self.patch_prepare_executor_instance()

        (download_directory / ignored_file).write_text("ignored content")

        self.assertFalse(self.task.configure_for_execution(download_directory))

        log_file_contents = (
            self.task.log_directory / "configure_for_execution.log"
        ).read_text()

        self.assertEqual(
            log_file_contents,
            f"No *.dsc, *.deb or *.udeb to be analyzed. "
            f"Files: ['{ignored_file}']\n",
        )

    def patch_extract_package_name_arch(self) -> MagicMock:
        """Patch self.task._extract_package_name_arch, return Mock."""
        patcher = mock.patch.object(
            self.task, "_extract_package_name_arch", autospec=True
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)

        return mocked

    def test_configure_for_execution_deb_and_udeb(self) -> None:
        """One .deb and one .udeb: self.task._lintian_targets has both."""
        directory = self.create_temporary_directory()

        file1 = self.create_temporary_file(suffix=".deb", directory=directory)
        file2 = self.create_temporary_file(suffix=".udeb", directory=directory)

        self.patch_extract_package_name_arch().return_value = None, None

        self.patch_prepare_executor_instance()

        success = self.task.configure_for_execution(directory)

        self.assertTrue(success)
        assert self.task._lintian_targets
        self.assertCountEqual(self.task._lintian_targets, [file1, file2])

    def test_configure_for_execution_create_mappings(self) -> None:
        """
        configure_for_execution() update member variables.

        Updates Lintian_architecture_to_packages,
        Lintian._package_name_to_filename.
        """
        directory = self.create_temporary_directory()

        # Create a binary package (.deb)
        binary_pkg_name = "hello"
        (
            binary_pkg_file := directory / f"{binary_pkg_name}_amd64.deb"
        ).write_text("test")

        # Create a source package (.dsc)
        source_pkg_file = directory / "hello_all.dsc"
        source_pkg_name = self.write_dsc_example_file(source_pkg_file)["Source"]
        source_pkg_name += " source"

        # Create a file that is ignored
        (ignored_file := directory / "package.tar.xz").write_text("not-used")

        extract_package_name_arch_patcher = mock.patch.object(
            self.task, "_extract_package_name_arch"
        )
        extract_file_name_mocked = extract_package_name_arch_patcher.start()
        filename_to_package_name_arch = {
            source_pkg_file.name: (source_pkg_name, "source"),
            ignored_file.name: (None, None),
            binary_pkg_file.name: (binary_pkg_name, "amd64"),
        }

        extract_file_name_mocked.side_effect = lambda file: (
            filename_to_package_name_arch.get(file.name)
        )
        self.addCleanup(extract_package_name_arch_patcher.stop)

        self.patch_prepare_executor_instance()

        self.task.configure_for_execution(directory)

        # self.task._package_name_to_filename is updated
        self.assertEqual(
            self.task._package_name_to_filename,
            {
                binary_pkg_name: binary_pkg_file.name,
                source_pkg_name: source_pkg_file.name,
            },
        )

        # self.task._architecture_to_packages is updated
        self.assertEqual(
            self.task._architecture_to_packages,
            {"source": {"hello source"}, "amd64": {"hello"}},
        )

    def patch_get_lintian_version(self, version: str) -> Any:
        """
        Patch Lintian._get_lintian_version().

        :param version: version returned by the mock
        :return: the patcher
        """
        patcher = mock.patch.object(Lintian, "_get_lintian_version")
        mocked = patcher.start()
        mocked.return_value = version
        self.addCleanup(patcher.stop)

        return patcher

    def test_configure_additional_properties_in_input(self) -> None:
        """Assert no additional properties in task_data input."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            self.configure_task(
                override={
                    "input": {
                        "source_artifact": 5,
                        "binary_artifacts": [],
                        "some_additional_property": 87,
                    },
                }
            )

    def test_build_consistency_no_errors(self) -> None:
        """There are no consistency errors."""
        build_directory = self.create_temporary_directory()
        (build_directory / Lintian.CAPTURE_OUTPUT_FILENAME).write_text("")
        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [],
        )

    def test_check_directory_for_consistency_errors(self) -> None:
        """There is one consistency error: no lintian.txt."""
        build_directory = self.create_temporary_directory()

        expected = f"{Lintian.CAPTURE_OUTPUT_FILENAME} not in {build_directory}"
        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [expected],
        )

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        exec_dir = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        lintian_output = exec_dir / self.task.CAPTURE_OUTPUT_FILENAME
        lintian_output.write_text("E: hello: invalid-file\n")

        # Set the analysis that will be uploaded

        summary_data = {
            "tags_count_by_severity": {"error": 1},
            "tags_found": ["invalid-file"],
            "overridden_tags_found": [],
            "lintian_version": "1.0.0",
            "distribution": "debian:unstable",
        }
        architecture_to_analysis = {
            "source": {
                "summary": {
                    **summary_data,
                    "package_filename": {"hello": "hello_1.0.dsc"},
                }
            },
            "all": {
                "summary": {
                    **summary_data,
                    "package_filename": {"hello-doc": "hello-doc_1.0_all.deb"},
                }
            },
            "amd64": {
                "summary": {
                    **summary_data,
                    "package_filename": {"hello": "hello_1.0_amd64.deb"},
                }
            },
        }

        create_architecture_to_analysis_patcher = mock.patch.object(
            self.task, "_create_architecture_to_analysis", autospec=True
        )
        create_architecture_to_analysis_mocked = (
            create_architecture_to_analysis_patcher.start()
        )
        create_architecture_to_analysis_mocked.return_value = (
            architecture_to_analysis
        )
        self.addCleanup(create_architecture_to_analysis_patcher.stop)

        # Used ty verify the relations
        self.task._source_artifacts_ids = [1]

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
            create_remote_artifact(id=11, workspace=workspace_name),
            create_remote_artifact(id=12, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for Lintian
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        self.task.upload_artifacts(
            exec_dir, execution_result=WorkRequestResults.SUCCESS
        )

        # Debusine Mock upload_artifact expected calls
        upload_artifact_calls = []
        for architecture, analysis in architecture_to_analysis.items():
            (
                analysis_file := exec_dir / f"analysis-{architecture}.json"
            ).write_text(json.dumps(analysis))
            lintian_artifact = LintianArtifact.create(
                analysis=analysis_file,
                lintian_output=lintian_output,
                architecture=architecture,
                summary=DebianLintianSummary.model_validate(
                    analysis["summary"]
                ),
            )

            upload_artifact_calls.append(
                call(
                    lintian_artifact,
                    workspace=workspace_name,
                    work_request=work_request_id,
                )
            )

        # Debusine mock relation_create expected calls
        relation_create_calls = []
        for uploaded_artifact, source_artifact_id in itertools.product(
            uploaded_artifacts, self.task._source_artifacts_ids
        ):
            relation_create_calls.append(
                call(uploaded_artifact.id, source_artifact_id, "relates-to")
            )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_has_calls(upload_artifact_calls)
        debusine_mock.relation_create.assert_has_calls(relation_create_calls)

    def test_get_lintian_version_raise_assertion(self) -> None:
        """
        _get_lintian_version() raise AssertionError.

        self.task.executor_instance is None.
        """
        self._get_lintian_version_patcher.stop()
        msg = (
            r"^Task\.executor_instance must be set before "
            r"calling _get_lintian_version\(\)$"
        )
        self.assertRaisesRegex(
            AssertionError, msg, self.task._get_lintian_version
        )

    def test_get_lintian_version(self) -> None:
        """
        _get_lintian_version() return the correct Lintian version.

        It also set self.task._lintian_version and use it if called again.
        """
        self._get_lintian_version_patcher.stop()
        version = "2.22.33"

        self.task.executor_instance = mock.Mock(spec=InstanceInterface)

        run_result = mock.Mock(spec=CompletedProcess)
        run_result.stdout = f"{version}\n"

        self.task.executor_instance.run.return_value = run_result

        self.assertEqual(self.task._get_lintian_version(), version)

        self.task.executor_instance.run.assert_called_with(
            ["lintian", "--print-version"],
            encoding="utf-8",
            errors="ignore",
            text=True,
            check=True,
        )

    def test_create_analysis(self) -> None:
        """_create_analysis() return a correct analysis."""
        lintian_output = (
            b"E: twitter-bootstrap3: alien-tag source-contains-empty-directory [debian/source/lintian-overrides:9]\n"  # noqa: E501
            b"W: jabber-muc source: obsolete-url-in-packaging https://download.gna.org/mu-conference/ [debian/control]\n"  # noqa: E501
            b"W: hello: some-warning\n"
            b"O: twitter-bootstrap3: tag-overridden\n"
            b"O: hello: tag-overridden ignored package hello not in analysis\n"
        )
        lintian_file = self.create_temporary_file(contents=lintian_output)

        packages = {"twitter-bootstrap3", "jabber-muc source"}

        self.task._package_name_to_filename = {
            "twitter-bootstrap3": "twitter-bootstrap3",
            "jabber-muc source": "jabber-muc.dsc",
        }

        parsed_output = self.task._parse_output(lintian_file)
        parsed_output.sort(
            key=lambda x: (
                x["package"],
                -self.task._severities_to_level[
                    x["severity"]
                ],  # negative to sort from highest to lowest
                x["tag"],
                x["note"],
            )
        )

        actual = self.task._create_analysis(parsed_output, packages)

        summary = {
            "tags_count_by_severity": {
                "error": 1,
                "warning": 1,
                "info": 0,
                "pedantic": 0,
                "experimental": 0,
                "overridden": 1,
                "classification": 0,
            },
            "package_filename": {
                "jabber-muc": "jabber-muc.dsc",
                "twitter-bootstrap3": "twitter-bootstrap3",
            },
            "tags_found": ["alien-tag", "obsolete-url-in-packaging"],
            "overridden_tags_found": ["tag-overridden"],
            "lintian_version": "1.0.0",
            "distribution": "debian:unstable",
        }

        # Tags from the parsed_output without "hello" (it is not being
        # analyzed) and without " source" in the package name
        normalized_tags = []

        for tag in parsed_output:
            if tag["package"] == "hello":
                continue

            normalized_tags.append(
                {**tag, "package": tag["package"].removesuffix(" source")}
            )

        expected = {
            "tags": normalized_tags,
            "summary": summary,
            "version": self.task.TASK_VERSION,
        }
        self.assertEqual(actual, expected)

    def run_fetch_input(
        self,
        input_override: dict[str, Any],
        dynamic_data: LintianDynamicData,
        expected_artifact_ids: list[int],
    ) -> None:
        """Verify that Lintian.fetch_input() downloads artifacts."""
        workspace = "System"

        self.configure_task(override={"input": input_override})
        self.task.work_request_id = 147
        self.task.workspace_name = "testing"
        self.task.dynamic_data = dynamic_data
        debusine_mock = self.mock_debusine()
        destination = self.create_temporary_directory()

        expected_calls = []
        artifact_responses = []
        expected_artifact_id_to_filenames = {}

        for artifact_id in expected_artifact_ids:
            expected_calls.append(
                call(
                    artifact_id, destination, tarball=False, workspace="testing"
                )
            )

            file_in_artifact = self.create_temporary_file()
            file_model = FileResponse(
                size=file_in_artifact.stat().st_size,
                checksums={"sha256": "not-used"},
                type="file",
                url="https://not-used",
            )

            filename = f"pkg-{artifact_id}_amd64.deb"

            artifact_responses.append(
                create_artifact_response(
                    id=artifact_id,
                    workspace=workspace,
                    category="Test",
                    created_at=dt.datetime.now(tz=dt.UTC),
                    data={},
                    download_tar_gz_url="https://example.com/not-used",
                    files=FilesResponseType({filename: file_model}),
                    files_to_upload=[],
                )
            )

            expected_artifact_id_to_filenames[artifact_id] = {filename}

        debusine_mock.download_artifact.side_effect = artifact_responses

        self.assertTrue(self.task.fetch_input(destination))

        # Artifacts were downloaded
        self.assertEqual(
            debusine_mock.download_artifact.call_args_list, expected_calls
        )

    def test_fetch_input_only_binary_artifacts(self) -> None:
        """Lintian.fetch_input() download binary_artifacts_ids."""
        artifacts = [1, 2]
        self.run_fetch_input(
            input_override={"binary_artifacts": artifacts},
            dynamic_data=LintianDynamicData(input_binary_artifacts_ids=[1, 2]),
            expected_artifact_ids=artifacts,
        )

    def test_fetch_input_only_source_artifact(self) -> None:
        """Lintian.fetch_input() download source_artifact_id."""
        artifact_id = 5
        self.run_fetch_input(
            input_override={"source_artifact": artifact_id},
            dynamic_data=LintianDynamicData(
                input_source_artifact_id=artifact_id
            ),
            expected_artifact_ids=[artifact_id],
        )

    def test_fetch_input_source_and_binary_artifact(self) -> None:
        """Lintian.fetch_input() download binary and source artifacts."""
        source_artifact_id = 5
        binary_artifacts_ids = [6, 7]
        self.run_fetch_input(
            input_override={
                "source_artifact": source_artifact_id,
                "binary_artifacts": binary_artifacts_ids,
            },
            dynamic_data=LintianDynamicData(
                input_source_artifact_id=source_artifact_id,
                input_binary_artifacts_ids=binary_artifacts_ids,
            ),
            expected_artifact_ids=[source_artifact_id, *binary_artifacts_ids],
        )

    def test_extract_package_name_arch_source_file(self) -> None:
        """Lintian._extract_package_name_arch() return pkgname + source."""
        file = self.create_temporary_file(suffix=".dsc")
        package_name = self.write_dsc_example_file(file)["Source"]

        self.assertEqual(
            self.task._extract_package_name_arch(file),
            (package_name + " source", "source"),
        )

    def test_extract_package_name_arch_raise_invalid_suffix(self) -> None:
        """Lintian._extract_package_name_arch() invalid suffix: None, None."""
        self.assertEqual(
            self.task._extract_package_name_arch(Path("name.tmp")), (None, None)
        )

    def test_extract_package_name_arch_deb_udeb(self) -> None:
        """Lintian._extract_package_name_arch() use DebFile correctly."""
        for architecture in ["amd64", "all"]:
            with self.subTest(architecture=architecture):
                filenames = [
                    "package-name_amd64.deb",
                    "package-name_amd64.udeb",
                ]
                package_name = "package-name"

                mock_control = MagicMock()
                mock_control.debcontrol.return_value = {
                    "Package": package_name,
                    "Architecture": architecture,
                }

                mock_deb_file_instance = MagicMock()
                mock_deb_file_instance.control = mock_control

                debfile_patcher = mock.patch(
                    "debusine.tasks.lintian.debfile.DebFile"
                )
                debfile_mocked = debfile_patcher.start()
                debfile_mocked.return_value = mock_deb_file_instance

                self.addCleanup(debfile_patcher.stop)

                for filename in filenames:
                    with self.subTest(filename=filename):
                        self.assertEqual(
                            self.task._extract_package_name_arch(
                                Path(filename)
                            ),
                            (package_name, architecture),
                        )

                        debfile_mocked.assert_called_with(Path(filename))

    def test_configure_for_execution_invalid_dsc(self) -> None:
        """Lintian._extract_package_name_arch() handles an invalid .dsc file."""
        directory = self.create_temporary_directory()
        dsc = directory / "hello.dsc"
        dsc.touch()

        self.assertEqual(
            self.task._extract_package_name_arch(dsc), (None, None)
        )

        log_file_contents = (
            self.task.log_directory / "configure_for_execution.log"
        ).read_text()
        self.assertEqual(
            log_file_contents, f"{dsc.name!r} does not contain Source field\n"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Add components to debian:system-tarball and debian:system-image artifacts."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps


def add_components(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Add components to artifacts."""
    Artifact = apps.get_model("db", "Artifact")

    for artifact in Artifact.objects.filter(
        category__in=("debian:system-tarball", "debian:system-image"),
        data__components__isnull=True,
    ):
        components = []
        if work_request := artifact.created_by_work_request:
            if repos := work_request.task_data.get(
                "bootstrap_repositories", []
            ):
                repo = repos[0]
                if "components" in repo:
                    components = repo["components"]
        artifact.data["components"] = components
        artifact.save()


def remove_components(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Remove components from artifacts."""
    Artifact = apps.get_model("db", "Artifact")

    for artifact in Artifact.objects.filter(
        category__in=("debian:system-tarball", "debian:system-image"),
        data__components__isnull=False,
    ):
        artifact.data.pop("components")
        artifact.save()


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0003_add_workspace_role_viewer"),
    ]
    replaces = [
        ("db", "0145_system_image_components"),
    ]

    operations = [
        migrations.RunPython(add_components, remove_components),
    ]
